// Copyright Notice
// ================
// BOCHS is Copyright 1994-1999 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
//
//     EMail:    bochs@world.std.com



#include "bochs.h"



// boundaries of consideration:
//
//  * physical memory boundary: 1024k (1Megabyte) (increments of...)
//  * A20 boundary:             1024k (1Megabyte)
//  * page boundary:            4k
//  * ROM boundary:             2k (dont care since we are only reading)
//  * segment boundary:         any



  void
BX_CPU_C::prefetch(void)
{
  // cs:eIP
  // prefetch QSIZE byte quantity aligned on corresponding boundary
  Bit32u new_linear_addr;
  Bit32u new_phy_addr;
  Bit32u temp_eip, temp_limit;

  temp_eip   = BX_CPU_THIS_PTR eip;
  temp_limit = BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled;

  new_linear_addr = BX_CPU_THIS_PTR cs.cache.u.segment.base + temp_eip;
  BX_CPU_THIS_PTR prev_linear_page = new_linear_addr & 0xfffff000;
  if (temp_eip > temp_limit) {
    bx_panic("prefetch: EIP > CS.limit\n");
    }

  if (BX_CPU_THIS_PTR cr0.pg) {
    // aligned block guaranteed to be all in one page, same A20 address
    new_phy_addr = itranslate_linear(new_linear_addr, CPL==3);
    new_phy_addr = A20ADDR(new_phy_addr);
    }
  else {
    new_phy_addr = A20ADDR(new_linear_addr);
    }

  if ( new_phy_addr >= BX_MEM.len ) {
    // don't take this out if dynamic translation enabled,
    // otherwise you must make a check to see if bytesleft is 0 after
    // a call to prefetch() in the dynamic code.
    bx_panic("prefetch: running in bogus memory\n");
    }

  // max physical address as confined by page boundary
  BX_CPU_THIS_PTR prev_phy_page = new_phy_addr & 0xfffff000;
  BX_CPU_THIS_PTR max_phy_addr = BX_CPU_THIS_PTR prev_phy_page | 0x00000fff;

  // check if segment boundary comes into play
  //if ((temp_limit - temp_eip) < 4096) {
  //  }

  BX_CPU_THIS_PTR bytesleft = (BX_CPU_THIS_PTR max_phy_addr - new_phy_addr) + 1;
  BX_CPU_THIS_PTR fetch_ptr = &BX_MEM.vector[new_phy_addr];
}


  // If control has transfered locally, it is possible the prefetch Q is
  // still valid.  This would happen for repeat instructions, and small
  // branches.
  void
BX_CPU_C::revalidate_prefetch_q(void)
{
  Bit32u new_linear_addr, new_linear_page, new_linear_offset;
  Bit32u new_phy_addr;

  new_linear_addr = BX_CPU_THIS_PTR cs.cache.u.segment.base + BX_CPU_THIS_PTR eip;

  new_linear_page = new_linear_addr & 0xfffff000;
  if (new_linear_page == BX_CPU_THIS_PTR prev_linear_page) {
    // same linear address, old linear->physical translation valid
    new_linear_offset = new_linear_addr & 0x00000fff;
    new_phy_addr = BX_CPU_THIS_PTR prev_phy_page | new_linear_offset;
    BX_CPU_THIS_PTR bytesleft = (BX_CPU_THIS_PTR max_phy_addr - new_phy_addr) + 1;
    BX_CPU_THIS_PTR fetch_ptr = &BX_MEM.vector[new_phy_addr];
    }
  else {
    BX_CPU_THIS_PTR bytesleft = 0; // invalidate prefetch Q
    }
}

  INLINE void
BX_CPU_C::invalidate_prefetch_q(void)
{
  BX_CPU_THIS_PTR bytesleft = 0;
}


  INLINE Bit8u
BX_CPU_C::fetch_next_byte(void)
{
  Bit8u ret8;

doit:
  if (BX_CPU_THIS_PTR bytesleft) {
    BX_INSTR_MEM_CODE(BX_CPU_THIS_PTR cs.cache.u.segment.base + EIP, 1);

    BX_CPU_THIS_PTR eip++;
    BX_CPU_THIS_PTR bytesleft--;
    ret8 = *BX_CPU_THIS_PTR fetch_ptr++;
    BX_INSTR_FETCH_BYTE(ret8);
    return( ret8 );
    }

  prefetch();
  goto doit;
}


  INLINE Bit16u
BX_CPU_C::fetch_next_word(void)
{
  Bit8u b_low, b_high;
  Bit16u ret16;

  BX_INSTR_MEM_CODE(BX_CPU_THIS_PTR cs.cache.u.segment.base + EIP, 2);

  if (BX_CPU_THIS_PTR bytesleft >= 2) {
    b_low  = *BX_CPU_THIS_PTR fetch_ptr++;
    b_high = *BX_CPU_THIS_PTR fetch_ptr++;
    BX_CPU_THIS_PTR eip += 2;
    BX_CPU_THIS_PTR bytesleft -= 2;
    ret16 = (b_high << 8) | b_low;
    BX_INSTR_FETCH_WORD(ret16);
    return( ret16 );
    }
  else if (BX_CPU_THIS_PTR bytesleft == 1) {
    b_low  = *BX_CPU_THIS_PTR fetch_ptr++;
    BX_CPU_THIS_PTR eip += 1;
    // BX_CPU_THIS_PTR bytesleft = 0;  // no need for this
    prefetch();
    b_high = *BX_CPU_THIS_PTR fetch_ptr++;
    BX_CPU_THIS_PTR eip += 1;
    BX_CPU_THIS_PTR bytesleft--;
    ret16 = (b_high<<8) | b_low;
    BX_INSTR_FETCH_WORD(ret16);
    return( ret16 );
    }
  else {
    prefetch();
    b_low  = *BX_CPU_THIS_PTR fetch_ptr++;
    b_high = *BX_CPU_THIS_PTR fetch_ptr++;
    BX_CPU_THIS_PTR eip += 2;
    BX_CPU_THIS_PTR bytesleft -= 2;
    ret16 = (b_high<<8) | b_low;
    BX_INSTR_FETCH_WORD(ret16);
    return( ret16 );
    }
}

  INLINE Bit32u
BX_CPU_C::fetch_next_dword(void)
{
  Bit32u ret32;

  BX_INSTR_MEM_CODE(BX_CPU_THIS_PTR cs.cache.u.segment.base + EIP, 4);

  if (BX_CPU_THIS_PTR bytesleft >= 4) {
    Bit8u b0, b1, b2, b3;

    b0 = *BX_CPU_THIS_PTR fetch_ptr++;
    b1 = *BX_CPU_THIS_PTR fetch_ptr++;
    b2 = *BX_CPU_THIS_PTR fetch_ptr++;
    b3 = *BX_CPU_THIS_PTR fetch_ptr++;
    BX_CPU_THIS_PTR eip += 4;
    BX_CPU_THIS_PTR bytesleft -= 4;
    ret32 = (b3<<24) | (b2<<16) | (b1<<8) | b0;
    BX_INSTR_FETCH_DWORD(ret32);
    return( ret32 );
    }
  else {
    Bit8u b[4];
    unsigned index;

    // code spans multiple pages
    for (index=0; index<4; index++) {
      if (BX_CPU_THIS_PTR bytesleft == 0)
        prefetch();
      b[index] = *BX_CPU_THIS_PTR fetch_ptr++;
      BX_CPU_THIS_PTR eip += 1;
      BX_CPU_THIS_PTR bytesleft--;
      }

    ret32 = (b[3]<<24) | (b[2]<<16) | (b[1]<<8) | b[0];
    BX_INSTR_FETCH_DWORD(ret32);
    return( ret32 );
    }
}
