// Copyright Notice
// ================
// BOCHS is Copyright 1994-1999 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
//
//     EMail:    bochs@world.std.com







#include "bochs.h"


  void
BX_CPU_C::UndefinedOpcode(BxInstruction_t *i)
{
  bx_printf("UndefinedOpcode: %02x causes exception 6\n",
            (unsigned) i->b1);
  exception(6, 0, 0);
}

  void
BX_CPU_C::NOP(BxInstruction_t *i)
{
}


  INLINE void
BX_CPU_C::HLT(BxInstruction_t *i)
{
  // hack to panic if HLT comes from BIOS
  if ( BX_CPU_THIS_PTR cs.selector.value == 0xf000 )
    bx_panic("HALT instruction encountered\n");

  if (CPL!=0) {
    bx_printf("HLT(): CPL!=0\n");
    exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }

  // stops instruction execution and places the processor in a
  // HALT state.  An enabled interrupt, NMI, or reset will resume
  // execution.  If interrupt is used to resume execution after HLT,
  // the saved CS:eIP points to instruction following HLT.
  while ( !(BX_INTR) ) {
    BX_TICK1();
    }
}

  INLINE void
BX_CPU_C::CLTS(BxInstruction_t *i)
{
#if BX_CPU_LEVEL < 2
  bx_panic("CLTS: not implemented for < 286\n");
#else

  if (v8086_mode()) bx_panic("clts: v8086 mode unsupported\n");

  /* read errata file */
  // does CLTS also clear NT flag???

  // #GP(0) if CPL is not 0
  if (CPL!=0) {
    bx_printf("CLTS(): CPL!=0\n");
    exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }

  BX_CPU_THIS_PTR cr0.ts = 0;
  BX_CPU_THIS_PTR cr0.val32 &= ~0x08;
  bx_printf("CLTS called\n");
#endif
}

  INLINE void
BX_CPU_C::INVD(BxInstruction_t *i)
{
  bx_printf("---------------\n");
  bx_printf("- INVD called -\n");
  bx_printf("---------------\n");
  invalidate_prefetch_q();
  BX_INSTR_CACHE_CNTRL(BX_INSTR_INVD);
}

  INLINE void
BX_CPU_C::WBINVD(BxInstruction_t *i)
{
  if (BX_CPU_THIS_PTR eflags.vm) bx_panic("clts: v8086 mode unsupported\n");
  if (BX_CPU_THIS_PTR cr0.pe) {
    if (CPL!=0) {
      bx_printf("WBINVD: CPL!=0\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      }
    }

  bx_printf("WBINVD: (ignoring)\n");
  invalidate_prefetch_q();
  BX_INSTR_CACHE_CNTRL(BX_INSTR_WBINVD);
}

  INLINE void
BX_CPU_C::INVLPG_Ea(BxInstruction_t *i)
{
  bx_panic("INVLPG_Ea:\n");
  invalidate_prefetch_q();
  BX_INSTR_TLB_CNTRL(BX_INSTR_INVLPG, 0);
}

  INLINE void
BX_CPU_C::MOV_DdRd(BxInstruction_t *i)
{
#if BX_CPU_LEVEL < 3
  bx_panic("MOV_DdRd: not supported on < 386\n");
#else
  Bit32u val_32;

  if (v8086_mode()) bx_panic("MOV_DdRd: v8086 mode unsupported\n");

  /* NOTES:
   *   32bit operands always used
   *   r/m field specifies general register
   *   mod field should always be 11 binary
   *   reg field specifies which special register
   *   Protected mode: #GP(0) if attempt to write a 1 to any reserved bit
   *     of CR4
   */

  if (i->mod != 0xc0) {
    bx_panic("MOV_DdRd(): rm field not a register!\n");
    }

  invalidate_prefetch_q();

  if (protected_mode() && CPL!=0) {
    bx_panic("MOV_DdRd: CPL!=0\n");
    /* #GP(0) if CPL is not 0 */
    exception(BX_GP_EXCEPTION, 0, 0);
    }

  val_32 = BX_READ_32BIT_REG(i->rm);
  if (bx_dbg.dreg)
    bx_printf("MOV_DdRd: DR[%u]=%08xh unhandled\n",
      (unsigned) i->nnn, (unsigned) val_32);

  switch (i->nnn) {
    case 0: // DR0
      BX_CPU_THIS_PTR dr0 = val_32;
      break;
    case 1: // DR1
      BX_CPU_THIS_PTR dr1 = val_32;
      break;
    case 2: // DR2
      BX_CPU_THIS_PTR dr2 = val_32;
      break;
    case 3: // DR3
      BX_CPU_THIS_PTR dr3 = val_32;
      break;
    case 4: // DR4
      BX_CPU_THIS_PTR dr4 = val_32;
      break;
    case 5: // DR5
      BX_CPU_THIS_PTR dr5 = val_32;
      break;
    case 6: // DR6
      BX_CPU_THIS_PTR dr6 = val_32;
      break;
    case 7: // DR7
      BX_CPU_THIS_PTR dr7 = val_32;
      // if any breakpointing is enabled, bail
      // looking for: GD, G3,L3,G2,L2,G1,L1,G0,L0
      if ( val_32 & 0x000020ff ) {
        bx_panic("MOV_DdRd: DR7=%08xh: enabling breakpointing\n", (unsigned) val_32);
        }
      break;
    default:
      bx_panic("MOV_DdRd: control register index out of range\n");
      break;
    }
#endif
}

  INLINE void
BX_CPU_C::MOV_RdDd(BxInstruction_t *i)
{
#if BX_CPU_LEVEL < 3
  bx_panic("MOV_RdDd: not supported on < 386\n");
#else
  Bit32u val_32;

  if (v8086_mode()) bx_panic("proc_ctrl: v8086 mode unsupported\n");

  /* NOTES:
   *   32bit operands always used
   *   r/m field specifies general register
   *   mod field should always be 11 binary
   *   reg field specifies which special register
   */

  if (i->mod != 0xc0) {
    bx_panic("MOV_RdDd(): rm field not a register!\n");
    }

  if (protected_mode() && CPL!=0) {
    bx_panic("MOV_RdDd: CPL!=0\n");
    /* #GP(0) if CPL is not 0 */
    exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }

  if (bx_dbg.dreg)
    bx_printf("MOV_RdDd: DR%u not implemented yet\n", i->nnn);

  switch (i->nnn) {
    case 0: // DR0
      val_32 = BX_CPU_THIS_PTR dr0;
      break;
    case 1: // DR1
      val_32 = BX_CPU_THIS_PTR dr1;
      break;
    case 2: // DR2
      val_32 = BX_CPU_THIS_PTR dr2;
      break;
    case 3: // DR3
      val_32 = BX_CPU_THIS_PTR dr3;
      break;
    case 4: // DR4
      val_32 = BX_CPU_THIS_PTR dr4;
      break;
    case 5: // DR5
      val_32 = BX_CPU_THIS_PTR dr5;
      break;
    case 6: // DR6
      val_32 = BX_CPU_THIS_PTR dr6;
      break;
    case 7: // DR7
      val_32 = BX_CPU_THIS_PTR dr7;
      break;
    default:
      bx_panic("MOV_RdDd: control register index out of range\n");
      val_32 = 0;
    }
  BX_WRITE_32BIT_REG(i->rm, val_32);
#endif
}


 INLINE void
BX_CPU_C::LMSW_Ew(BxInstruction_t *i)
{
#if BX_CPU_LEVEL < 2
  bx_panic("LMSW_Ew(): not supported on 8086!\n");
#else
  Bit16u msw;
  Boolean prev_pe, pe, mp, em, ts;

  if (v8086_mode()) bx_panic("proc_ctrl: v8086 mode unsupported\n");

  if ( protected_mode() ) {
    if ( CPL != 0 ) {
      bx_printf("LMSW: CPL != 0, CPL=%u\n", (unsigned) CPL);
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }


  if (i->mod == 0xc0) {
    msw = BX_READ_16BIT_REG(i->rm);
    }
  else {
    read_virtual_word(i->seg_reg, i->rm_addr, &msw);
    }

  /* LMSW does not affect PG,CD,NW,AM,WP,NE,ET bits, and cannot clear PE */
  prev_pe = BX_CPU_THIS_PTR cr0.pe;
  pe = msw & 0x01; msw >>= 1;
  mp = msw & 0x01; msw >>= 1;
  em = msw & 0x01; msw >>= 1;
  ts = msw & 0x01;

  if (mp || ts) {
    bx_panic("unsupported setting of MP, EM, TS bits in LMSW instructions\n");
    }
  BX_CPU_THIS_PTR cr0.em = em;
  BX_CPU_THIS_PTR cr0.val32 = (BX_CPU_THIS_PTR cr0.val32 & ~0x04) | (em<<2);

  if ((prev_pe==0) && pe) {
    // switching to protected mode
    BX_CPU_THIS_PTR cr0.pe = 1;
    BX_CPU_THIS_PTR cr0.val32 |= 0x01;
    enter_protected_mode();
    }

  else if (prev_pe && (pe==0)) {
    bx_panic("LMSW_Ew: transition to real mode not supported!\n");
    }
#endif /* BX_CPU_LEVEL < 2 */
}

 INLINE void
BX_CPU_C::SMSW_Ew(BxInstruction_t *i)
{
#if BX_CPU_LEVEL < 2
  bx_panic("SMSW_Ew: not supported yet!\n");
#else
  Bit16u msw;


#if BX_CPU_LEVEL == 2
  msw = 0xfff0; /* 80286 init value */
  msw |= (BX_CPU_THIS_PTR cr0.ts << 3) |
         (BX_CPU_THIS_PTR cr0.em << 2) |
         (BX_CPU_THIS_PTR cr0.mp << 1) |
         BX_CPU_THIS_PTR cr0.pe;
#else /* 386+ */
  /* reserved bits 0 ??? */
  /* should NE bit be included here ??? */
  // should ET bit be included here (AW)
  msw =  (BX_CPU_THIS_PTR cr0.ts << 3) |
         (BX_CPU_THIS_PTR cr0.em << 2) |
         (BX_CPU_THIS_PTR cr0.mp << 1) |
         BX_CPU_THIS_PTR cr0.pe;
#endif


  if (i->mod == 0xc0) {
    if (i->os_32) {
      BX_WRITE_32BIT_REG(i->rm, msw);  // zeros out high 16bits
      }
    else {
      BX_WRITE_16BIT_REG(i->rm, msw);
      }
    }
  else {
    write_virtual_word(i->seg_reg, i->rm_addr, &msw);
    }

#endif
}


  INLINE void
BX_CPU_C::MOV_CdRd(BxInstruction_t *i)
{
  // mov general register data to control register
#if BX_CPU_LEVEL < 3
  bx_panic("MOV_CdRd: not supported on < 386\n");
#else
  Bit32u val_32;


  if (v8086_mode()) bx_panic("proc_ctrl: v8086 mode unsupported\n");

  /* NOTES:
   *   32bit operands always used
   *   r/m field specifies general register
   *   mod field should always be 11 binary
   *   reg field specifies which special register
   *   Protected mode: #GP(0) if attempt to write a 1 to any reserved bit
   *     of CR4
   */

  if (i->mod != 0xc0) {
    bx_panic("MOV_CdRd(): rm field not a register!\n");
    }

  invalidate_prefetch_q();

  if (protected_mode() && CPL!=0) {
    bx_panic("MOV_CdRd: CPL!=0\n");
    /* #GP(0) if CPL is not 0 */
    exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }

  val_32 = BX_READ_32BIT_REG(i->rm);

  switch (i->nnn) {
    case 0: // CR0 (MSW)
      // bx_printf("MOV_CdRd:CR0: R32 = %08x\n @CS:EIP %04x:%04x ",
      //   (unsigned) val_32,
      //   (unsigned) BX_CPU_THIS_PTR cs.selector.value,
      //   (unsigned) BX_CPU_THIS_PTR eip);
      SetCR0(val_32);
      break;

    case 1: /* CR1 */
      bx_panic("MOV_CdRd: CR1 not implemented yet\n");
      break;
    case 2: /* CR2 */
      bx_printf("MOV_CdRd: CR2 not implemented yet\n");
      if (bx_dbg.creg)
        bx_printf("MOV_CdRd: CR2 = reg\n");
      BX_CPU_THIS_PTR cr2 = val_32;
      break;
    case 3: /* CR3 */
      if (bx_dbg.creg)
        bx_printf("MOV_CdRd:(%08x)\n", (unsigned) val_32);
      CR3_change(val_32);
      BX_INSTR_TLB_CNTRL(BX_INSTR_MOV_CR3, val_32);
      break;
    case 4: /* CR4 */
      bx_panic("MOV_CdRd: CR4 not implemented yet\n");
      break;
    default:
      bx_panic("MOV_CdRd: control register index out of range\n");
      break;
    }
#endif
}

  INLINE void
BX_CPU_C::MOV_RdCd(BxInstruction_t *i)
{
  // mov control register data to register
#if BX_CPU_LEVEL < 3
  bx_panic("MOV_RdCd: not supported on < 386\n");
#else
  Bit32u val_32;

  if (v8086_mode()) bx_panic("proc_ctrl: v8086 mode unsupported\n");

  /* NOTES:
   *   32bit operands always used
   *   r/m field specifies general register
   *   mod field should always be 11 binary
   *   reg field specifies which special register
   */

  if (i->mod != 0xc0) {
    bx_panic("MOV_RdCd(): rm field not a register!\n");
    }

  if (protected_mode() && CPL!=0) {
    bx_panic("MOV_RdCd: CPL!=0\n");
    /* #GP(0) if CPL is not 0 */
    exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }

  switch (i->nnn) {
    case 0: // CR0 (MSW)
      val_32 = BX_CPU_THIS_PTR cr0.val32;
#if 0
      bx_printf("MOV_RdCd:CR0: R32 = %08x\n @CS:EIP %04x:%04x\n",
        (unsigned) val_32,
        (unsigned) BX_CPU_THIS_PTR cs.selector.value,
        (unsigned) BX_CPU_THIS_PTR eip);
#endif
      break;
    case 1: /* CR1 */
      bx_panic("MOV_RdCd: CR1 not implemented yet\n");
      val_32 = 0;
      break;
    case 2: /* CR2 */
      if (bx_dbg.creg)
        bx_printf("MOV_RdCd: CR2\n");
      val_32 = BX_CPU_THIS_PTR cr2;
      break;
    case 3: /* CR3 */
      if (bx_dbg.creg)
        bx_printf("MOV_RdCd: reading CR3\n");
      val_32 = BX_CPU_THIS_PTR cr3;
      break;
    case 4: /* CR4 */
      bx_panic("MOV_RdCd: CR4 not implemented yet\n");
      val_32 = 0;
      break;
    default:
      bx_panic("MOV_RdCd: control register index out of range\n");
      val_32 = 0;
    }
  BX_WRITE_32BIT_REG(i->rm, val_32);
#endif
}

void BX_CPU_C::MOV_TdRd(BxInstruction_t *i) {bx_panic("MOV_TdRd:\n");}
void BX_CPU_C::MOV_RdTd(BxInstruction_t *i) {bx_panic("MOV_RdTd:\n");}

  INLINE void
BX_CPU_C::WAIT(BxInstruction_t *i)
{
#if BX_CPU_LEVEL < 3
  bx_panic("WAIT: not implemented for < 386\n");
#elif BX_CPU_LEVEL > 3
  bx_panic("WAIT: not implemented for > 386\n");
#else // BX_CPU_LEVEL == 3

  BX_CPU_THIS_PTR single_step_event = 0; // ??? WAIT doesn't generate single steps

  if ( BX_CPU_THIS_PTR cr0.ts && BX_CPU_THIS_PTR cr0.mp ) {
    exception(BX_NM_EXCEPTION, 0, 0); // no error
    }
#endif
}

  INLINE void
BX_CPU_C::LOADALL(BxInstruction_t *i)
{
#if BX_CPU_LEVEL < 2
  bx_panic("undocumented LOADALL instruction not supported on 8086\n");
#else
  Bit16u msw, tr, flags, ip, ldtr;
  Bit16u ds_raw, ss_raw, cs_raw, es_raw;
  Bit16u di, si, bp, sp, bx, dx, cx, ax;
  Bit16u base_15_0, limit;
  Bit8u  base_23_16, access;

  if (v8086_mode()) bx_panic("proc_ctrl: v8086 mode unsupported\n");

#if BX_CPU_LEVEL > 2
  bx_panic("loadall: not implemented for 386\n");
  /* ??? need to set G and other bits, and compute .limit_scaled also */
  /* for all segments CS,DS,SS,... */
#endif

  if (BX_CPU_THIS_PTR cr0.pe) {
    bx_panic(
      "LOADALL not yet supported for protected mode\n");
    }

bx_panic("LOADALL: handle CR0.val32\n");
  /* MSW */
  BX_MEM.read_physical(0x806, 2, &msw);
  BX_CPU_THIS_PTR cr0.pe = (msw & 0x01); msw >>= 1;
  BX_CPU_THIS_PTR cr0.mp = (msw & 0x01); msw >>= 1;
  BX_CPU_THIS_PTR cr0.em = (msw & 0x01); msw >>= 1;
  BX_CPU_THIS_PTR cr0.ts = (msw & 0x01);

  //bx_printf("LOADALL: pe=%u, mp=%u, em=%u, ts=%u\n",
  //  (unsigned) BX_CPU_THIS_PTR cr0.pe, (unsigned) BX_CPU_THIS_PTR cr0.mp,
  //  (unsigned) BX_CPU_THIS_PTR cr0.em, (unsigned) BX_CPU_THIS_PTR cr0.ts);

  if (BX_CPU_THIS_PTR cr0.pe || BX_CPU_THIS_PTR cr0.mp || BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts)
    bx_panic("LOADALL set PE, MP, EM or TS bits in MSW!\n");

  /* TR */
  BX_MEM.read_physical(0x816, 2, &tr);
  BX_CPU_THIS_PTR tr.selector.value = tr;
  BX_CPU_THIS_PTR tr.selector.rpl   = (tr & 0x03);  tr >>= 2;
  BX_CPU_THIS_PTR tr.selector.ti    = (tr & 0x01);  tr >>= 1;
  BX_CPU_THIS_PTR tr.selector.index = tr;
  BX_MEM.read_physical(0x860, 2, &base_15_0);
  BX_MEM.read_physical(0x862, 1, &base_23_16);
  BX_MEM.read_physical(0x863, 1, &access);
  BX_MEM.read_physical(0x864, 2, &limit);


  BX_CPU_THIS_PTR tr.cache.valid =
  BX_CPU_THIS_PTR tr.cache.p           = (access & 0x80) >> 7;
  BX_CPU_THIS_PTR tr.cache.dpl         = (access & 0x60) >> 5;
  BX_CPU_THIS_PTR tr.cache.segment     = (access & 0x10) >> 4;
  BX_CPU_THIS_PTR tr.cache.type        = (access & 0x0f);
  BX_CPU_THIS_PTR tr.cache.u.tss286.base  = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR tr.cache.u.tss286.limit = limit;

  if ( (BX_CPU_THIS_PTR tr.selector.value & 0xfffc) == 0 ) {
    BX_CPU_THIS_PTR tr.cache.valid = 0;
    }
  if ( BX_CPU_THIS_PTR tr.cache.valid == 0 ) {
    }
  if ( BX_CPU_THIS_PTR tr.cache.u.tss286.limit < 43 ) {
    BX_CPU_THIS_PTR tr.cache.valid = 0;
    }
  if ( BX_CPU_THIS_PTR tr.cache.type != 1 ) {
    BX_CPU_THIS_PTR tr.cache.valid = 0;
    }
  if ( BX_CPU_THIS_PTR tr.cache.segment ) {
    BX_CPU_THIS_PTR tr.cache.valid = 0;
    }
  if (BX_CPU_THIS_PTR tr.cache.valid==0) {
    BX_CPU_THIS_PTR tr.cache.u.tss286.base   = 0;
    BX_CPU_THIS_PTR tr.cache.u.tss286.limit  = 0;
    BX_CPU_THIS_PTR tr.cache.p            = 0;
    BX_CPU_THIS_PTR tr.selector.value     = 0;
    BX_CPU_THIS_PTR tr.selector.index     = 0;
    BX_CPU_THIS_PTR tr.selector.ti        = 0;
    BX_CPU_THIS_PTR tr.selector.rpl       = 0;
    }


  /* FLAGS */
  BX_MEM.read_physical(0x818, 2, &flags);
  write_flags(flags, 1, 1);

  /* IP */
  BX_MEM.read_physical(0x81a, 2, &ip);
  IP = ip;

  /* LDTR */
  BX_MEM.read_physical(0x81c, 2, &ldtr);
  BX_CPU_THIS_PTR ldtr.selector.value = ldtr;
  BX_CPU_THIS_PTR ldtr.selector.rpl   = (ldtr & 0x03);  ldtr >>= 2;
  BX_CPU_THIS_PTR ldtr.selector.ti    = (ldtr & 0x01);  ldtr >>= 1;
  BX_CPU_THIS_PTR ldtr.selector.index = ldtr;
  if ( (BX_CPU_THIS_PTR ldtr.selector.value & 0xfffc) == 0 ) {
    BX_CPU_THIS_PTR ldtr.cache.valid   = 0;
    BX_CPU_THIS_PTR ldtr.cache.p       = 0;
    BX_CPU_THIS_PTR ldtr.cache.segment = 0;
    BX_CPU_THIS_PTR ldtr.cache.type    = 0;
    BX_CPU_THIS_PTR ldtr.cache.u.ldt.base = 0;
    BX_CPU_THIS_PTR ldtr.cache.u.ldt.limit = 0;
    BX_CPU_THIS_PTR ldtr.selector.value = 0;
    BX_CPU_THIS_PTR ldtr.selector.index = 0;
    BX_CPU_THIS_PTR ldtr.selector.ti    = 0;
    }
  else {
    BX_MEM.read_physical(0x854, 2, &base_15_0);
    BX_MEM.read_physical(0x856, 1, &base_23_16);
    BX_MEM.read_physical(0x857, 1, &access);
    BX_MEM.read_physical(0x858, 2, &limit);
    BX_CPU_THIS_PTR ldtr.cache.valid      =
    BX_CPU_THIS_PTR ldtr.cache.p          = access >> 7;
    BX_CPU_THIS_PTR ldtr.cache.dpl        = (access >> 5) & 0x03;
    BX_CPU_THIS_PTR ldtr.cache.segment    = (access >> 4) & 0x01;
    BX_CPU_THIS_PTR ldtr.cache.type       = (access & 0x0f);
    BX_CPU_THIS_PTR ldtr.cache.u.ldt.base = (base_23_16 << 16) | base_15_0;
    BX_CPU_THIS_PTR ldtr.cache.u.ldt.limit = limit;

    if (access == 0) {
      bx_panic("loadall: LDTR case access byte=0.\n");
      }
    if ( BX_CPU_THIS_PTR ldtr.cache.valid==0 ) {
      bx_panic("loadall: ldtr.valid=0\n");
      }
    if (BX_CPU_THIS_PTR ldtr.cache.segment) { /* not a system segment */
      bx_printf("         AR byte = %02x\n", (unsigned) access);
      bx_panic("loadall: LDTR descriptor cache loaded with non system segment\n");
      }
    if ( BX_CPU_THIS_PTR ldtr.cache.type != 2 ) {
      bx_panic("loadall: LDTR.type(%u) != 2\n", (unsigned) (access & 0x0f));
      }
    }

  /* DS */
  BX_MEM.read_physical(0x81e, 2, &ds_raw);
  BX_CPU_THIS_PTR ds.selector.value = ds_raw;
  BX_CPU_THIS_PTR ds.selector.rpl   = (ds_raw & 0x03);  ds_raw >>= 2;
  BX_CPU_THIS_PTR ds.selector.ti    = (ds_raw & 0x01);  ds_raw >>= 1;
  BX_CPU_THIS_PTR ds.selector.index = ds_raw;
  BX_MEM.read_physical(0x848, 2, &base_15_0);
  BX_MEM.read_physical(0x84a, 1, &base_23_16);
  BX_MEM.read_physical(0x84b, 1, &access);
  BX_MEM.read_physical(0x84c, 2, &limit);
  BX_CPU_THIS_PTR ds.cache.u.segment.base = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR ds.cache.u.segment.limit = limit;
  BX_CPU_THIS_PTR ds.cache.u.segment.a          = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ds.cache.u.segment.r_w        = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ds.cache.u.segment.c_ed       = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ds.cache.u.segment.executable = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ds.cache.segment    = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ds.cache.dpl        = (access & 0x03); access >>= 2;
  BX_CPU_THIS_PTR ds.cache.valid      =
  BX_CPU_THIS_PTR ds.cache.p          = (access & 0x01);

  if ( (BX_CPU_THIS_PTR ds.selector.value & 0xfffc) == 0 ) {
    BX_CPU_THIS_PTR ds.cache.valid = 0;
    }
  if (BX_CPU_THIS_PTR ds.cache.valid==0  ||
      BX_CPU_THIS_PTR ds.cache.segment==0) {
    bx_panic("loadall: DS invalid\n");
    }

  /* SS */
  BX_MEM.read_physical(0x820, 2, &ss_raw);
  BX_CPU_THIS_PTR ss.selector.value = ss_raw;
  BX_CPU_THIS_PTR ss.selector.rpl   = (ss_raw & 0x03); ss_raw >>= 2;
  BX_CPU_THIS_PTR ss.selector.ti    = (ss_raw & 0x01); ss_raw >>= 1;
  BX_CPU_THIS_PTR ss.selector.index = ss_raw;
  BX_MEM.read_physical(0x842, 2, &base_15_0);
  BX_MEM.read_physical(0x844, 1, &base_23_16);
  BX_MEM.read_physical(0x845, 1, &access);
  BX_MEM.read_physical(0x846, 2, &limit);
  BX_CPU_THIS_PTR ss.cache.u.segment.base = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR ss.cache.u.segment.limit = limit;
  BX_CPU_THIS_PTR ss.cache.u.segment.a          = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ss.cache.u.segment.r_w        = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ss.cache.u.segment.c_ed       = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ss.cache.u.segment.executable = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ss.cache.segment    = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ss.cache.dpl        = (access & 0x03); access >>= 2;
  BX_CPU_THIS_PTR ss.cache.p          = (access & 0x01);

  if ( (BX_CPU_THIS_PTR ss.selector.value & 0xfffc) == 0 ) {
    BX_CPU_THIS_PTR ss.cache.valid = 0;
    }
  if (BX_CPU_THIS_PTR ss.cache.valid==0  ||
      BX_CPU_THIS_PTR ss.cache.segment==0) {
    bx_panic("loadall: SS invalid\n");
    }


  /* CS */
  BX_MEM.read_physical(0x822, 2, &cs_raw);
  BX_CPU_THIS_PTR cs.selector.value = cs_raw;
  BX_CPU_THIS_PTR cs.selector.rpl   = (cs_raw & 0x03); cs_raw >>= 2;

  //bx_printf("LOADALL: setting cs.selector.rpl to %u\n",
  //  (unsigned) BX_CPU_THIS_PTR cs.selector.rpl);

  BX_CPU_THIS_PTR cs.selector.ti    = (cs_raw & 0x01); cs_raw >>= 1;
  BX_CPU_THIS_PTR cs.selector.index = cs_raw;
  BX_MEM.read_physical(0x83c, 2, &base_15_0);
  BX_MEM.read_physical(0x83e, 1, &base_23_16);
  BX_MEM.read_physical(0x83f, 1, &access);
  BX_MEM.read_physical(0x840, 2, &limit);
  BX_CPU_THIS_PTR cs.cache.u.segment.base = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR cs.cache.u.segment.limit = limit;
  BX_CPU_THIS_PTR cs.cache.u.segment.a          = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR cs.cache.u.segment.r_w        = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR cs.cache.u.segment.c_ed       = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR cs.cache.u.segment.executable = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR cs.cache.segment    = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR cs.cache.dpl        = (access & 0x03); access >>= 2;
  BX_CPU_THIS_PTR cs.cache.p          = (access & 0x01);

  if ( (BX_CPU_THIS_PTR cs.selector.value & 0xfffc) == 0 ) {
    BX_CPU_THIS_PTR cs.cache.valid = 0;
    }
  if (BX_CPU_THIS_PTR cs.cache.valid==0  ||
      BX_CPU_THIS_PTR cs.cache.segment==0) {
    bx_panic("loadall: CS invalid\n");
    }

  /* ES */
  BX_MEM.read_physical(0x824, 2, &es_raw);
  BX_CPU_THIS_PTR es.selector.value = es_raw;
  BX_CPU_THIS_PTR es.selector.rpl   = (es_raw & 0x03); es_raw >>= 2;
  BX_CPU_THIS_PTR es.selector.ti    = (es_raw & 0x01); es_raw >>= 1;
  BX_CPU_THIS_PTR es.selector.index = es_raw;
  BX_MEM.read_physical(0x836, 2, &base_15_0);
  BX_MEM.read_physical(0x838, 1, &base_23_16);
  BX_MEM.read_physical(0x839, 1, &access);
  BX_MEM.read_physical(0x83a, 2, &limit);
  BX_CPU_THIS_PTR es.cache.u.segment.base = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR es.cache.u.segment.limit = limit;
  BX_CPU_THIS_PTR es.cache.u.segment.a          = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR es.cache.u.segment.r_w        = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR es.cache.u.segment.c_ed       = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR es.cache.u.segment.executable = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR es.cache.segment    = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR es.cache.dpl        = (access & 0x03); access >>= 2;
  BX_CPU_THIS_PTR es.cache.p          = (access & 0x01);

#if 0
    bx_printf("cs.dpl = %02x\n", (unsigned) BX_CPU_THIS_PTR cs.cache.dpl);
    bx_printf("ss.dpl = %02x\n", (unsigned) BX_CPU_THIS_PTR ss.cache.dpl);
    bx_printf("BX_CPU_THIS_PTR ds.dpl = %02x\n", (unsigned) BX_CPU_THIS_PTR ds.cache.dpl);
    bx_printf("BX_CPU_THIS_PTR es.dpl = %02x\n", (unsigned) BX_CPU_THIS_PTR es.cache.dpl);
    bx_printf("LOADALL: setting cs.selector.rpl to %u\n",
      (unsigned) BX_CPU_THIS_PTR cs.selector.rpl);
    bx_printf("LOADALL: setting ss.selector.rpl to %u\n",
      (unsigned) BX_CPU_THIS_PTR ss.selector.rpl);
    bx_printf("LOADALL: setting ds.selector.rpl to %u\n",
      (unsigned) BX_CPU_THIS_PTR ds.selector.rpl);
    bx_printf("LOADALL: setting es.selector.rpl to %u\n",
      (unsigned) BX_CPU_THIS_PTR es.selector.rpl);
#endif

  if ( (BX_CPU_THIS_PTR es.selector.value & 0xfffc) == 0 ) {
    BX_CPU_THIS_PTR es.cache.valid = 0;
    }
  if (BX_CPU_THIS_PTR es.cache.valid==0  ||
      BX_CPU_THIS_PTR es.cache.segment==0) {
    bx_panic("loadall: ES invalid\n");
    }

  /* DI */
  BX_MEM.read_physical(0x826, 2, &di);
  DI = di;

  /* SI */
  BX_MEM.read_physical(0x828, 2, &si);
  SI = si;

  /* BP */
  BX_MEM.read_physical(0x82a, 2, &bp);
  BP = bp;

  /* SP */
  BX_MEM.read_physical(0x82c, 2, &sp);
  SP = sp;

  /* BX */
  BX_MEM.read_physical(0x82e, 2, &bx);
  BX = bx;

  /* DX */
  BX_MEM.read_physical(0x830, 2, &dx);
  DX = dx;

  /* CX */
  BX_MEM.read_physical(0x832, 2, &cx);
  CX = cx;

  /* AX */
  BX_MEM.read_physical(0x834, 2, &ax);
  AX = ax;

  /* GDTR */
  BX_MEM.read_physical(0x84e, 2, &base_15_0);
  BX_MEM.read_physical(0x850, 1, &base_23_16);
  BX_MEM.read_physical(0x851, 1, &access);
  BX_MEM.read_physical(0x852, 2, &limit);
  BX_CPU_THIS_PTR gdtr.base = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR gdtr.limit = limit;

#if 0
  if (access)
      bx_printf("LOADALL: GDTR access bits not 0 (%02x).\n",
        (unsigned) access);
#endif

  /* IDTR */
  BX_MEM.read_physical(0x85a, 2, &base_15_0);
  BX_MEM.read_physical(0x85c, 1, &base_23_16);
  BX_MEM.read_physical(0x85d, 1, &access);
  BX_MEM.read_physical(0x85e, 2, &limit);
  BX_CPU_THIS_PTR idtr.base = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR idtr.limit = limit;
#endif
}


  INLINE void
BX_CPU_C::CPUID(BxInstruction_t *i)
{
  invalidate_prefetch_q();
#if BX_CPUID_HACK
  //
  // stick hack here
  //
  bx_panic("CPUID: not implemented\n");
#else
  bx_panic("CPUID: not implemented\n");
#endif
}

  void
BX_CPU_C::SetCR0(Bit32u val_32)
{
  // from either MOV_CdRd() or debug functions
  // protection checks made already or forcing from debug
  Boolean prev_pe, prev_pg;

  prev_pe = BX_CPU_THIS_PTR cr0.pe;
  prev_pg = BX_CPU_THIS_PTR cr0.pg;

  BX_CPU_THIS_PTR cr0.pe = val_32 & 0x01;
  BX_CPU_THIS_PTR cr0.mp = (val_32 >> 1) & 0x01;
  BX_CPU_THIS_PTR cr0.em = (val_32 >> 2) & 0x01;
  BX_CPU_THIS_PTR cr0.ts = (val_32 >> 3) & 0x01;
  // cr0.et is hardwired to 1
#if BX_CPU_LEVEL >= 4
  BX_CPU_THIS_PTR cr0.ne = (val_32 >> 5)  & 0x01;
  BX_CPU_THIS_PTR cr0.wp = (val_32 >> 16) & 0x01;
  BX_CPU_THIS_PTR cr0.am = (val_32 >> 18) & 0x01;
  BX_CPU_THIS_PTR cr0.nw = (val_32 >> 29) & 0x01;
  BX_CPU_THIS_PTR cr0.cd = (val_32 >> 30) & 0x01;
#endif
  BX_CPU_THIS_PTR cr0.pg = (val_32 >> 31) & 0x01;

  // handle reserved bits behaviour
#if BX_CPU_LEVEL == 3
  BX_CPU_THIS_PTR cr0.val32 = val_32 | 0x7ffffff0;
#elif BX_CPU_LEVEL == 5
  BX_CPU_THIS_PTR cr0.val32 = val_32 | 0x00000010;
#elif BX_CPU_LEVEL == 6
  BX_CPU_THIS_PTR cr0.val32 = (val_32 | 0x00000010) & 0xe005003f;
#else
#error "MOV_CdRd: implement reserved bits behaviour for this CPU_LEVEL"
#endif

  if (BX_CPU_THIS_PTR cr0.ts)
    bx_printf("MOV_CdRd:CR0.TS set 0x%x\n", (unsigned) val_32);

  if (prev_pe==0 && BX_CPU_THIS_PTR cr0.pe) {
    enter_protected_mode();
    }
  else if (prev_pe==1 && BX_CPU_THIS_PTR cr0.pe==0) {
    enter_real_mode();
    }

  if (prev_pg==0 && BX_CPU_THIS_PTR cr0.pg)
    enable_paging();
  else if (prev_pg==1 && BX_CPU_THIS_PTR cr0.pg==0)
    disable_paging();
}
