// Copyright Notice
// ================
// BOCHS is Copyright 1994-1999 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
//
//     EMail:    bochs@world.std.com



extern "C" {
#include <signal.h>
}

#include "bochs.h"

static unsigned doit = 0;


static void bx_dbg_usage(void);
static char bx_debug_rc_fname[BX_MAX_PATH];
static char tmp_buf[512];
static char *tmp_buf_ptr;
static char *argv0 = NULL;


#if BX_NUM_SIMULATORS >= 2
#define BX_DBG_IO_JOURNAL_SIZE        1024
#define BX_DBG_UCMEM_JOURNAL_SIZE     1024
#define BX_DBG_ASYNC_JOURNAL_SIZE     1024
#define BX_DBG_MASTER_MODE              10
#define BX_DBG_SLAVE_MODE               11
#define BX_DBG_DEFAULT_ICOUNT_QUANTUM   50

static unsigned bx_dbg_cosimulateN(bx_dbg_icount_t count);
static Boolean  bx_dbg_compare_sim_iaddr(void);
static Boolean  bx_dbg_compare_sim_cpu(void);
static Boolean  bx_dbg_compare_sim_memory(void);
static void     bx_dbg_journal_a20_event(unsigned val);
#endif


static struct {
#if BX_NUM_SIMULATORS >= 2
  // some fields used only for cosimulation
  unsigned icount_quantum;
  unsigned master_slave_mode;
  unsigned master, slave;
  struct {
    struct {
      Bit8u    op;
      Bit8u    len;
      Bit16u   addr;
      Bit32u   value;
      } element[BX_DBG_IO_JOURNAL_SIZE];
    unsigned size;
    unsigned head, tail;
    } IO_journal;

  struct {
    struct {
      Bit8u    op;
      Bit8u    len;
      Bit32u   addr;
      Bit32u   value;
      } element[BX_DBG_UCMEM_JOURNAL_SIZE];
    unsigned size;
    unsigned head, tail;
    } UCmem_journal;

// need to handle DMA stuff in here...

#define BX_DBG_ASYNC_JOURNAL_NONE   0
#define BX_DBG_ASYNC_JOURNAL_A20    1
#define BX_DBG_ASYNC_JOURNAL_IAC    2
#define BX_DBG_ASYNC_JOURNAL_NMI    3
#define BX_DBG_ASYNC_JOURNAL_RESET  4

  // Asynchronous events at the boundaries they are *taken* by the master simulator.
  // These are replayed back to the slave at the same boundaries.
  struct {
    struct {
      unsigned        what; // A20, INTR, NMI, RESET, IAC, ...
      bx_dbg_icount_t icount;
      union {
        struct {
          unsigned val;
          } a20, nmi, reset, iac;
        // perhaps other more complex types here
        } u;
      } element[BX_DBG_ASYNC_JOURNAL_SIZE];
    unsigned size;
    unsigned head, tail;
    } async_journal;

  struct {
    Boolean iaddr;
    Boolean cpu;
    Boolean memory;
    } compare_at_sync;
#endif  // #if BX_NUM_SIMULATORS >= 2

  // some fields used for single CPU debugger
  Boolean  auto_disassemble;
  unsigned disassemble_size;
  char     default_display_format;
  char     default_unit_size;
  Bit32u   default_addr;
  unsigned next_bpoint_id;

  // last icount known to be in sync
#if BX_DBG_ICOUNT_SIZE == 32
  Bit32u last_sync_icount;
#else  // BX_DBG_ICOUNT_SIZE == 64
  Bit64u last_sync_icount;
#endif
  } bx_debugger;



// cosim commands for handling of comparison of simulator
// environments when both simulators have reached a common
// point (synchronized).

// cosim compare_at_sync iaddr  (default is on)
// cosim compare_at_sync cpu    (default is off)
// cosim compare_at_sync memory (default is off)
// cosim compare iaddr
// cosim compare cpu
// cosim compare memory


typedef struct {
  FILE    *fp;
  char     fname[BX_MAX_PATH];
  unsigned lineno;
  } bx_infile_stack_entry_t;

bx_infile_stack_entry_t bx_infile_stack[BX_INFILE_DEPTH];
int                     bx_infile_stack_index = 0;

static int  bx_nest_infile(char *path);

static void bx_debug_ctrlc_handler(int signum);


static void bx_unnest_infile(void);
static void bx_get_command(void);
static void bx_dbg_print_guard_results(void);
static void bx_dbg_breakpoint_changed(void);

bx_dbg_callback_t bx_dbg_callback[BX_NUM_SIMULATORS];
bx_guard_t        bx_guard;
bx_guard_found_t  bx_guard_found[BX_NUM_SIMULATORS];


// DMA stuff
void bx_dbg_post_dma_reports(void);
#define BX_BATCH_DMA_BUFSIZE 512

static struct {
  unsigned this_many;  // batch this many max before posting events
  unsigned Qsize;      // this many have been batched
  struct {
    Bit32u   addr;   // address of DMA op
    unsigned len;    // number of bytes in op
    unsigned what;   // BX_READ or BX_WRITE
    Bit32u   val;    // value of DMA op
    bx_dbg_icount_t icount; // icount at this dma op
    } Q[BX_BATCH_DMA_BUFSIZE];
  } bx_dbg_batch_dma;


// some buffers for disassembly
#if BX_DISASM
static Bit8u bx_disasm_ibuf[32];
static char  bx_disasm_tbuf[512];
#endif



  int
bx_dbg_main(int argc, char *argv[])
{
  int i, bochs_argc=0, sim1_argc=0, sim2_argc=0;
  char **bochs_argv = NULL;
  char **sim1_argv = NULL;
  char **sim2_argv = NULL;

  bx_dbg_batch_dma.this_many = 1;
  bx_dbg_batch_dma.Qsize     = 0;

  // initialize callback functions, and guard environment
  memset(bx_dbg_callback, 0, sizeof(bx_dbg_callback));
  memset(&bx_guard, 0, sizeof(bx_guard));
  bx_guard.async.irq = 1;
  bx_guard.async.dma = 1;
  memset(bx_guard_found, 0, sizeof(bx_guard_found));

  memset(&bx_debugger, 0, sizeof(bx_debugger));
#if BX_NUM_SIMULATORS >= 2
  bx_debugger.icount_quantum = BX_DBG_DEFAULT_ICOUNT_QUANTUM;
  bx_debugger.IO_journal.size = 0;
  bx_debugger.IO_journal.head = 0;
  bx_debugger.IO_journal.tail = 0;
  bx_debugger.UCmem_journal.size = 0;
  bx_debugger.UCmem_journal.head = 0;
  bx_debugger.UCmem_journal.tail = 0;
  bx_debugger.async_journal.size = 0;
  bx_debugger.async_journal.head = 0;
  bx_debugger.async_journal.tail = 0;
  bx_debugger.master = 0;
  bx_debugger.slave  = 1;
  bx_debugger.compare_at_sync.iaddr  = 1;
  bx_debugger.compare_at_sync.cpu    = 0;
  bx_debugger.compare_at_sync.memory = 0;
#endif
  bx_debugger.auto_disassemble = 1;
  bx_debugger.disassemble_size = 32;
  bx_debugger.default_display_format = 'x';
  bx_debugger.default_unit_size      = 'w';
  bx_debugger.default_addr = 0;
  bx_debugger.next_bpoint_id = 1;
  bx_debugger.last_sync_icount = 0;


  argv0 = strdup(argv[0]);

  bx_debug_rc_fname[0] = '\0';

  bochs_argv = (char **) &argv[0];
  sim1_argv = bochs_argv;  // start out with something reasonable
  sim2_argv = bochs_argv;  // start out with something reasonable
  bochs_argc = 1;
  sim1_argc = 1;
  sim2_argc = 1;

  // process "-rc pathname" option, if it exists
  i = 1;
  if ( (argc >= 2) && !strcmp(argv[1], "-rc") ) {
    if ( argc == 2 ) {
      fprintf(stderr, "%s: -rc option used, but no path specified.\n",
        argv[0]);
      bx_dbg_usage();
      exit(1);
      }
    strncpy(bx_debug_rc_fname, argv[2], BX_MAX_PATH-1);
    i += 2; // skip past "-rc" and filename
    bochs_argv = (char **) &argv[2];
    }

  // process options to bochs framework
  for (; i<argc; i++) {
    if (strcmp(argv[i], "-sim1") == 0) {
      break;
      }
    else if (strcmp(argv[i], "-sim2") == 0) {
      break;
      }
    bochs_argc++;
    }

  if (i<argc) {  // more args to process
    // process options to each CPU simulator
    if (strcmp(argv[i], "-sim1") == 0) {
process_sim1:
      sim1_argv = (char **) &argv[i];
      i++;
      for (; i<argc; i++) {
        if (strcmp(argv[i], "-sim2") == 0)
          goto process_sim2;
        sim1_argc++;
        }
      }
    else if (strcmp(argv[i], "-sim2") == 0) {
process_sim2:
      sim2_argv = (char **) &argv[i];
      i++;
      for (; i<argc; i++) {
        if (strcmp(argv[i], "-sim1") == 0)
          goto process_sim1;
        sim2_argc++;
        }
      }
    }


  bx_infile_stack_index = 0;
  bx_infile_stack[0].fp = stdin;
  strncpy(bx_infile_stack[0].fname, argv[0], BX_MAX_PATH);
  bx_infile_stack[0].fname[BX_MAX_PATH-1] = 0;
  bx_infile_stack[0].lineno = 0;


  if (bx_debug_rc_fname[0] == '\0') {
    fprintf(stderr, "%s: Warning: no rc file specified.\n", argv[0]);
    }
  else {
    fprintf(stderr, "%s: using rc file '%s'.\n", argv[0], bx_debug_rc_fname);
    // if there's an error, the user will know about it before proceeding
    (void) bx_nest_infile(bx_debug_rc_fname);
    }

#if BX_DISASM
  memset(bx_disasm_ibuf, 0, sizeof(bx_disasm_ibuf));
#endif

  BX_SIM1_INIT(&bx_dbg_callback[0], sim1_argc, sim1_argv);
#if BX_NUM_SIMULATORS > 1
  BX_SIM2_INIT(&bx_dbg_callback[1], sim2_argc, sim2_argv);
#endif

  // call init routine for overall framework and IO device setup
  bx_bochs_init(bochs_argc, bochs_argv);

  // call init routines for each CPU+mem simulator
  bx_dbg_callback[0].reset_cpu(BX_RESET_HARDWARE);
  bx_dbg_callback[0].init_mem(bx_options.memory.megs * 1024*1024);
  bx_dbg_callback[0].load_ROM(bx_options.rom.path, bx_options.rom.address);
  bx_dbg_callback[0].load_ROM(bx_options.vgarom.path, 0xc0000);
#if BX_NUM_SIMULATORS > 1
  bx_dbg_callback[1].reset_cpu(BX_RESET_HARDWARE);
  bx_dbg_callback[1].init_mem(bx_options.memory.megs * 1024*1024);
  bx_dbg_callback[1].load_ROM(bx_options.rom.path, bx_options.rom.address);
  bx_dbg_callback[1].load_ROM(bx_options.vgarom.path, 0xc0000);
#endif

  // setup Ctrl-C handler
  signal(SIGINT, bx_debug_ctrlc_handler);

  bx_dbg_user_input_loop();

  bx_dbg_exit(0);
  return(0); // keep compiler happy
}


  void
bx_dbg_usage(void)
{
  fprintf(stderr, "usage: %s [-rc path] [-sim1 ... ] [-sim2 ... ]\n", argv0);
}


  void
bx_dbg_user_input_loop(void)
{
  int reti;
  unsigned include_cmd_len = strlen(BX_INCLUDE_CMD);

  while ( 1 ) {
    bx_get_command();
    if ( (*tmp_buf_ptr == '\n') || (*tmp_buf_ptr == 0) ) {
      if (bx_infile_stack_index == 0)
        fprintf(stderr, "\n");
      }
    else if ( (strncmp(tmp_buf_ptr, BX_INCLUDE_CMD, include_cmd_len) == 0) &&
              (tmp_buf_ptr[include_cmd_len] == ' ' ||
               tmp_buf_ptr[include_cmd_len] == '\t') ) {
      char *ptr;
      int len;

      ptr = tmp_buf_ptr + include_cmd_len+1;
      while ( *ptr==' ' || *ptr=='\t' )
        ptr++;

      len = strlen(ptr);
      if (len == 0) {
        fprintf(stderr, "%s: no filename given to 'source' command.\n", argv0);
        if (bx_infile_stack_index > 0) {
          fprintf(stderr, "%s: ERROR in source file causes exit.\n", argv0);
          bx_dbg_exit(1);
          }
        continue;
        }
      ptr[len-1] = 0; // get rid of newline
      reti = bx_nest_infile(ptr);
      if ((reti==0) && (bx_infile_stack_index > 0)) {
        fprintf(stderr, "%s: ERROR in source file causes exit.\n", argv0);
        bx_dbg_exit(1);
        }
      }
    else {
      // Give a chance to the command line extensions, to
      // consume the command.  If they return 0, then
      // we need to process the command.  A return of 1
      // means, the extensions have handled the command
      if ( bx_dbg_extensions(tmp_buf_ptr)==0 ) {
        // process command here
        bx_add_lex_input(tmp_buf_ptr);
        yyparse();
        }
      }
    }
}


  void
bx_get_command(void)
{
  char *charptr_ret;

  bx_infile_stack[bx_infile_stack_index].lineno++;

  if (bx_infile_stack_index == 0) {
    fprintf(stderr, "<bochs:%d> ", bx_infile_stack[bx_infile_stack_index].lineno);
    }

  charptr_ret = fgets(tmp_buf, 512,
    bx_infile_stack[bx_infile_stack_index].fp);
  if (charptr_ret == NULL) {
    // see if error was due to EOF condition
    if (feof(bx_infile_stack[bx_infile_stack_index].fp)) {
      if (bx_infile_stack_index > 0) {
        // nested level of include files, pop back to previous one
        bx_unnest_infile();
        }
      else {
        // not nested, sitting at stdin prompt, user wants out
        bx_dbg_quit_command();
        }

      // call recursively
      bx_get_command();
      return;
      }

    // error was not EOF, see if it was from a Ctrl-C
    if (bx_guard.interrupt_requested) {
      tmp_buf[0] = '\n';
      tmp_buf[1] = 0;
      tmp_buf_ptr = &tmp_buf[0];
      bx_guard.interrupt_requested = 0;
      return;
      }

    fprintf(stderr, "fgets() returned ERROR.\n");
fprintf(stderr, "intr request was %u\n", bx_guard.interrupt_requested);
    bx_dbg_exit(1);
    }
  tmp_buf_ptr = &tmp_buf[0];

  // look for first non-whitespace character
  while ( ((*tmp_buf_ptr == ' ') || (*tmp_buf_ptr == '\t')) &&
          (*tmp_buf_ptr != '\n') && (*tmp_buf_ptr != 0) ) {
    tmp_buf_ptr++;
    }

  return;
}

  int
bx_nest_infile(char *path)
{
  FILE *tmp_fp;

  tmp_fp = fopen(path, "r");
  if (!tmp_fp) {
    fprintf(stderr, "%s: can not open file '%s' for reading.\n",
            argv0, path);
    return(0);
    }

  if ( (bx_infile_stack_index+1) >= BX_INFILE_DEPTH ) {
    fprintf(stderr, "%s: source files nested too deeply\n", argv0);
    return(0);
    }

  bx_infile_stack_index++;
  bx_infile_stack[bx_infile_stack_index].fp = tmp_fp;
  strncpy(bx_infile_stack[bx_infile_stack_index].fname, path, BX_MAX_PATH);
  bx_infile_stack[bx_infile_stack_index].fname[BX_MAX_PATH-1] = 0;
  bx_infile_stack[bx_infile_stack_index].lineno = 0;
  return(1);
}

  void
bx_unnest_infile(void)
{
  if (bx_infile_stack_index <= 0) {
    fprintf(stderr, "%s: ERROR: unnest_infile(): nesting level = 0.\n",
      argv0);
    bx_dbg_exit(1);
    }

  fclose(bx_infile_stack[bx_infile_stack_index].fp);
  bx_infile_stack_index--;
}

  int
yywrap(void)
{
  fprintf(stderr, "%s: ERROR: yywrap() called.\n", argv0);
  bx_dbg_exit(1);
  return(0); // keep compiler quiet
}


  void
yyerror(char *s)
{
  fprintf(stderr, "%s:%d: %s at '%s'\n",
    bx_infile_stack[bx_infile_stack_index].fname,
    bx_infile_stack[bx_infile_stack_index].lineno,
    s, yytext);

  if (bx_infile_stack_index > 0) {
    fprintf(stderr, "%s: ERROR in source file causes exit.\n", argv0);
    bx_dbg_exit(1);
    }
}

  void
bx_debug_ctrlc_handler(int signum)
{
  UNUSED(signum);

  signal(SIGINT, bx_debug_ctrlc_handler);
  bx_guard.interrupt_requested = 1;
}

  void
bx_dbg_exit(int code)
{
  fprintf(stderr, "before sim1_exit\n");
  bx_dbg_callback[0].atexit();

#if BX_NUM_SIMULATORS >= 2
  fprintf(stderr, "before sim2_exit\n");
  bx_dbg_callback[1].atexit();
#endif

  bx_atexit();

  fprintf(stderr, "before exit\n");
  exit(code);
}


//
// comands invoked from parser
//

  void
bx_dbg_quit_command(void)
{
  fprintf(stderr, "Quit\n");
  bx_dbg_exit(0);
}

  void
bx_dbg_continue_command(void)
{
  // continue executing, until a guard found

#if BX_NUM_SIMULATORS >= 2
  bx_guard.interrupt_requested = 0;
  while (1) {
    if ( !bx_dbg_cosimulateN(bx_debugger.icount_quantum) )
      break;
    }
#else
  // run indefinitely
  bx_guard.icount = 0;
  bx_guard.guard_for &= ~BX_DBG_GUARD_ICOUNT; // not looking for icount
  bx_guard.guard_for |=  BX_DBG_GUARD_CTRL_C; // stop on Ctrl-C

  bx_guard_found[0].guard_found = 0;
  bx_guard_found[0].icount = 0;

  bx_guard.interrupt_requested = 0;
  bx_dbg_callback[0].execute();
#endif

  BX_INSTR_DEBUG_PROMPT();
  bx_dbg_print_guard_results();
}

  void
bx_dbg_stepN_command(bx_dbg_icount_t count)
{
  if (count == 0) {
    fprintf(stderr, "Error: stepN: count=0\n");
    return;
    }

#if BX_NUM_SIMULATORS >= 2
  bx_guard.interrupt_requested = 0;
  bx_dbg_cosimulateN(count);
#else
  // single CPU
  bx_guard.icount = count;
  bx_guard.guard_for |= BX_DBG_GUARD_ICOUNT; // looking for icount
  bx_guard.guard_for |= BX_DBG_GUARD_CTRL_C; // or Ctrl-C

  bx_guard_found[0].guard_found = 0;
  bx_guard_found[0].icount = 0;

  bx_guard.interrupt_requested = 0;
  bx_dbg_callback[0].execute();
#endif

  BX_INSTR_DEBUG_PROMPT();
  bx_dbg_print_guard_results();
}


#if BX_NUM_SIMULATORS >= 2
  unsigned
bx_dbg_cosimulateN(bx_dbg_icount_t count)
{
  // execute both master & slave for count instructions,
  // handling asynchronous events, etc.
  // returns 0 = didn't get through all count instructions
  //             either a guard was hit, or a divergence occurred
  //         1 = got through all count instructions

  unsigned master, slave;
  bx_dbg_icount_t master_icount, slave_icount;
  Boolean bail_out = 0;
  unsigned ret = 0;
  Boolean save_INTR;
  Boolean pre_A20, post_A20;
  unsigned async_head;
  bx_dbg_icount_t async_icount, curr_icount;

  if (count == 0) {
    fprintf(stderr, "Error: cosimulateN: count=0\n");
    bx_dbg_exit(1);
    }

  bx_guard.guard_for |= BX_DBG_GUARD_ICOUNT;  // stop at icount
  bx_guard.guard_for &= ~BX_DBG_GUARD_CTRL_C; // ignore Ctrl-C

one_time_slice:
  // take minimum of requested count and maximum count quantum
  if (count > bx_debugger.icount_quantum)
    bx_guard.icount = bx_debugger.icount_quantum;
  else
    bx_guard.icount = count;

  // for now, assume...
  master = bx_debugger.master;
  slave  = bx_debugger.slave;

  // run master simulator
  bx_debugger.master_slave_mode = BX_DBG_MASTER_MODE;
  if (bx_guard.interrupt_requested) {
    bail_out = 1;
fprintf(stderr, "ctrlc typed\n");
    }
  bx_guard_found[master].guard_found = 0;
  bx_guard_found[master].icount = 0;
if (doit) fprintf(stderr, "requesting run of master for %u\n",
  (unsigned) bx_guard.icount);
  // save A20 value before master run
  pre_A20  = bx_pc_system.get_enable_a20();

  bx_dbg_callback[master].execute();
  post_A20  = bx_pc_system.get_enable_a20(); // A20 after master run
  master_icount = bx_guard_found[master].icount;
  slave_icount = 0;
  if (master_icount)
    bx_pc_system.tickn(master_icount);
  save_INTR = bx_pc_system.INTR; // value after master run
  bx_pc_system.INTR = 0; // in case slave uses directly
  // Change A20 for slave run to model what it was at beginning of
  // master run, only if it needs to be changed.
  if (pre_A20 != post_A20) {
    bx_pc_system.set_enable_a20(pre_A20);
    if (bx_dbg_callback[slave].set_A20)
      bx_dbg_callback[slave].set_A20(pre_A20);
    }


  // if guard was anything except for icount, we should terminate
  // after synchronizing slave to master
  if (bx_guard_found[master].guard_found & ~BX_DBG_GUARD_ICOUNT)
    bail_out = 1;


  // Synchronize slave to master.  Account for Ctrl-C's typed during execution of
  // slave.
  bx_debugger.master_slave_mode = BX_DBG_SLAVE_MODE;
  do {
    // run slave for remaining instructions to catch up to master
    curr_icount = master_icount - slave_icount;
    if (bx_debugger.async_journal.size) {
      // If there were asynchronous events which occurred while the
      // master was running, have to run the slave up to each of these
      // points individually, and force it to take them on exactly the
      // same boundaries.
      async_head   = bx_debugger.async_journal.head;
      async_icount = bx_debugger.async_journal.element[async_head].icount;
      curr_icount  = async_icount;  // only run to next async event
      }
    else {
      async_head   = 0; // keep compiler happy
      async_icount = 0; // keep compiler happy
      }

    bx_guard_found[slave].guard_found = 0;
    bx_guard_found[slave].icount = 0;
    bx_guard.icount = curr_icount;

    if (curr_icount) {
      // Async event may be before completion of any instructions,
      // for example taking of interrupt.
if (doit) fprintf(stderr, "requesting run of slave for %u\n",
  (unsigned) bx_guard.icount);
      bx_dbg_callback[slave].execute();
      }
    slave_icount += bx_guard_found[slave].icount;
    if (bx_guard_found[slave].guard_found & ~BX_DBG_GUARD_ICOUNT) {
      bail_out = 1;
      // If user type Ctrl-C we're done after synchronizing.  If not,
      // then we have reached a true guard, and it's time to bail.
      if (bx_guard_found[slave].guard_found &
          ~(BX_DBG_GUARD_ICOUNT | BX_DBG_GUARD_CTRL_C))
        break;
      }
    if (bx_debugger.async_journal.size) {
      // sanity check: slave should be at async point
      if (bx_guard_found[slave].icount != async_icount) {
        fprintf(stderr, "Error: comsimulateN: async: slave not at sync point.\n");
        bx_dbg_exit(1);
        }
      switch (bx_debugger.async_journal.element[async_head].what) {
        case BX_DBG_ASYNC_JOURNAL_IAC:
if (doit)
  fprintf(stderr, "slave: forcing interrupt %u\n",
    bx_debugger.async_journal.element[async_head].u.iac.val);

          bx_dbg_callback[slave].force_interrupt(
            bx_debugger.async_journal.element[async_head].u.iac.val);
          break;
        case BX_DBG_ASYNC_JOURNAL_A20:
          bx_pc_system.set_enable_a20(
            bx_debugger.async_journal.element[async_head].u.a20.val);
          if (bx_dbg_callback[slave].set_A20)
            bx_dbg_callback[slave].set_A20(
              bx_debugger.async_journal.element[async_head].u.a20.val);
          break;
        case BX_DBG_ASYNC_JOURNAL_NMI:
        case BX_DBG_ASYNC_JOURNAL_RESET:
        default:
          fprintf(stderr, "Error: cosimulateN: unimplemented async event.\n");
        }
      // async event processed, dequeue it
      bx_debugger.async_journal.size--;
      bx_debugger.async_journal.head++;
      }
    } while (slave_icount < master_icount);

  bx_pc_system.INTR = save_INTR; // restore INTR to value after master run

  // At this point, both simulators should be at the same point.  Either
  // they have finished executing for the desired count, or at least for
  // a time quantum.  Check to see if the environments are in sync.
  if (bx_debugger.compare_at_sync.iaddr && bx_dbg_compare_sim_iaddr())
      bail_out = 1;
  else if (bx_debugger.compare_at_sync.cpu && bx_dbg_compare_sim_cpu())
      bail_out = 1;
  else if (bx_debugger.compare_at_sync.memory && bx_dbg_compare_sim_memory())
      bail_out = 1;

  if (bail_out) {
    ret = 0; // didn't complete, stopped
    }
  else {
    count -= master_icount;
    // last icount known to be in sync
    bx_debugger.last_sync_icount += master_icount;
    if (count)
      goto one_time_slice;
    ret = 1; // completed OK
    }

  bx_guard.guard_for &= ~BX_DBG_GUARD_ICOUNT;
  return(ret);
}
#endif  // #if BX_NUM_SIMULATORS >= 2


#if BX_NUM_SIMULATORS >= 2
  Boolean
bx_dbg_compare_sim_iaddr(void)
{
  // returns 0 = same, 1 = different
  if (bx_guard_found[0].laddr != bx_guard_found[1].laddr) {

    fprintf(stderr,
#if BX_DBG_ICOUNT_SIZE == 32
      "Divergence: last know synchronized icount was %lu\n",
      (unsigned long) bx_debugger.last_sync_icount
#else  // BX_DBG_ICOUNT_SIZE == 64
      "Divergence: last know synchronized icount was %Lu\n",
      (unsigned long long) bx_debugger.last_sync_icount
#endif
      );

//    fprintf(stderr, "Divergence: sim[0].laddr=%x, sim[1].laddr=%x\n",
//      (unsigned) bx_guard_found[0].laddr,
//      (unsigned) bx_guard_found[1].laddr);
    return(1); // different
    }
  return(0); // same
}

  Boolean
bx_dbg_compare_sim_cpu(void)
{
  fprintf(stderr, "compare_sim_iaddr: not finished\n");
  return(0);
}

  Boolean
bx_dbg_compare_sim_memory(void)
{
  fprintf(stderr, "compare_sim_iaddr: not finished\n");
  return(0);
}
#endif



  void
bx_dbg_print_guard_results(void)
{
  unsigned i;
  unsigned sim;

for (sim=0; sim<BX_NUM_SIMULATORS; sim++) {
  if (bx_guard_found[sim].guard_found & BX_DBG_GUARD_ICOUNT) {
    }
  else if (bx_guard_found[sim].guard_found & BX_DBG_GUARD_CTRL_C) {
    }
#if BX_DBG_SUPPORT_VIR_BPOINT
  else if (bx_guard_found[sim].guard_found & BX_DBG_GUARD_IADDR_VIR) {
    i = bx_guard_found[sim].iaddr_index;
    fprintf(stderr, "(%u) Breakpoint %u, 0x%x (0x%x:0x%x)\n",
            sim,
            bx_guard.iaddr.vir[i].bpoint_id,
            bx_guard_found[sim].laddr,
            bx_guard_found[sim].cs,
            bx_guard_found[sim].eip);
    }
#endif
#if BX_DBG_SUPPORT_LIN_BPOINT
  else if (bx_guard_found[sim].guard_found & BX_DBG_GUARD_IADDR_LIN) {
    i = bx_guard_found[sim].iaddr_index;
    fprintf(stderr, "(%u) Breakpoint %u, 0x%x in ?? ()\n",
            sim,
            bx_guard.iaddr.lin[i].bpoint_id,
            bx_guard_found[sim].laddr);
    }
#endif
#if BX_DBG_SUPPORT_PHY_BPOINT
  else if (bx_guard_found[sim].guard_found & BX_DBG_GUARD_IADDR_PHY) {
    i = bx_guard_found[sim].iaddr_index;
    fprintf(stderr, "(%u) Breakpoint %u, 0x%x in ?? ()\n",
            sim,
            bx_guard.iaddr.phy[i].bpoint_id,
            bx_guard_found[sim].laddr);
    }
#endif
  else {
    fprintf(stderr, "Error: (%u) print_guard_results: guard_found ?\n", sim);
    }

 fprintf(stderr, "(%u) 0x%x in ?? ()\n", sim,
            bx_guard_found[sim].laddr);


#if BX_DISASM
  if (bx_debugger.auto_disassemble) {
    Bit32u phy;
    Boolean valid;

    bx_dbg_callback[sim].xlate_linear2phy(bx_guard_found[sim].laddr, &phy, &valid);

    if (valid) {
      unsigned ilen;

      bx_dbg_callback[sim].getphymem(phy, 16, bx_disasm_ibuf);
      ilen = bx_disassemble.disasm(bx_guard_found[sim].is_32bit_code,
                                   bx_disasm_ibuf, bx_disasm_tbuf);

      if (bx_guard_found[sim].is_32bit_code) {
        fprintf(stderr, "(%u) %04x:%08x: ", sim,
                (unsigned) bx_guard_found[sim].cs,
                (unsigned) bx_guard_found[sim].eip);
        }
      else {
        fprintf(stderr, "(%u) %04x:%04x: ", sim,
                (unsigned) bx_guard_found[sim].cs,
                (unsigned) bx_guard_found[sim].eip);
        }
      for (unsigned j=0; j<ilen; j++)
        fprintf(stderr, "%02x", (unsigned) bx_disasm_ibuf[j]);
      fprintf(stderr, ": %s\n", bx_disasm_tbuf);
      }
    else {
      fprintf(stderr, "(%u) ??? (physical address not available)\n", sim);
      }
    }
#endif  // #if BX_DISASM
  }
}


  void
bx_dbg_breakpoint_changed(void)
{
#if BX_DBG_SUPPORT_VIR_BPOINT
  if (bx_guard.iaddr.num_virtual)
    bx_guard.guard_for |= BX_DBG_GUARD_IADDR_VIR;
  else
    bx_guard.guard_for &= ~BX_DBG_GUARD_IADDR_VIR;
#endif

#if BX_DBG_SUPPORT_LIN_BPOINT
  if (bx_guard.iaddr.num_linear)
    bx_guard.guard_for |= BX_DBG_GUARD_IADDR_LIN;
  else
    bx_guard.guard_for &= ~BX_DBG_GUARD_IADDR_LIN;
#endif

#if BX_DBG_SUPPORT_PHY_BPOINT
  if (bx_guard.iaddr.num_physical)
    bx_guard.guard_for |= BX_DBG_GUARD_IADDR_PHY;
  else
    bx_guard.guard_for &= ~BX_DBG_GUARD_IADDR_PHY;
#endif
}

  void
bx_dbg_del_breakpoint_command(unsigned handle)
{
  unsigned i;

#if BX_DBG_SUPPORT_VIR_BPOINT
  // see if breakpoint is a virtual breakpoint
  for (i=0; i<bx_guard.iaddr.num_virtual; i++) {
    if (bx_guard.iaddr.vir[i].bpoint_id == handle) {
      // found breakpoint, delete it by shifting remaining entries left
      for (unsigned j=i; j<(bx_guard.iaddr.num_virtual-1); j++) {
        bx_guard.iaddr.vir[j] = bx_guard.iaddr.vir[j+1];
        }
      bx_guard.iaddr.num_virtual--;
      goto done;
      }
    }
#endif

#if BX_DBG_SUPPORT_LIN_BPOINT
  // see if breakpoint is a linear breakpoint
  for (i=0; i<bx_guard.iaddr.num_linear; i++) {
    if (bx_guard.iaddr.lin[i].bpoint_id == handle) {
      // found breakpoint, delete it by shifting remaining entries left
      for (unsigned j=i; j<(bx_guard.iaddr.num_linear-1); j++) {
        bx_guard.iaddr.lin[j] = bx_guard.iaddr.lin[j+1];
        }
      bx_guard.iaddr.num_linear--;
      goto done;
      }
    }
#endif

#if BX_DBG_SUPPORT_PHY_BPOINT
  // see if breakpoint is a physical breakpoint
  for (i=0; i<bx_guard.iaddr.num_physical; i++) {
    if (bx_guard.iaddr.phy[i].bpoint_id == handle) {
      // found breakpoint, delete it by shifting remaining entries left
      for (unsigned j=i; j<(bx_guard.iaddr.num_physical-1); j++) {
        bx_guard.iaddr.phy[j] = bx_guard.iaddr.phy[j+1];
        }
      bx_guard.iaddr.num_physical--;
      goto done;
      }
    }
#endif

  fprintf(stderr, "Error: breakpoint %u not found.\n", handle);
  return;

done:
  bx_dbg_breakpoint_changed();
}

  void
bx_dbg_vbreakpoint_command(Boolean specific, Bit32u cs, Bit32u eip)
{
#if BX_DBG_SUPPORT_VIR_BPOINT
  if (specific == 0) {
    fprintf(stderr, "Error: vbreak without address not implemented yet.\n");
    return;
    }

  if (bx_guard.iaddr.num_virtual >= BX_DBG_MAX_VIR_BPOINTS) {
    fprintf(stderr, "Error: no more virtual breakpoint slots left.\n");
    fprintf(stderr, "Error: see BX_DBG_MAX_VIR_BPOINTS.\n");
    return;
    }

  bx_guard.iaddr.vir[bx_guard.iaddr.num_virtual].cs  = cs;
  bx_guard.iaddr.vir[bx_guard.iaddr.num_virtual].eip = eip;
  bx_guard.iaddr.vir[bx_guard.iaddr.num_virtual].bpoint_id = bx_debugger.next_bpoint_id++;
  bx_guard.iaddr.num_virtual++;
  bx_guard.guard_for |= BX_DBG_GUARD_IADDR_VIR;

#else
  fprintf(stderr, "Error: virtual breakpoint support not compiled in.\n");
  fprintf(stderr, "Error: see BX_DBG_SUPPORT_VIR_BPOINT.\n");
#endif
}

  void
bx_dbg_lbreakpoint_command(Boolean specific, Bit32u laddress)
{
#if BX_DBG_SUPPORT_LIN_BPOINT
  if (specific == 0) {
    fprintf(stderr, "Error: lbreak without address not implemented yet.\n");
    return;
    }

  if (bx_guard.iaddr.num_linear >= BX_DBG_MAX_LIN_BPOINTS) {
    fprintf(stderr, "Error: no more linear breakpoint slots left.\n");
    fprintf(stderr, "Error: see BX_DBG_MAX_LIN_BPOINTS.\n");
    return;
    }

  bx_guard.iaddr.lin[bx_guard.iaddr.num_linear].addr = laddress;
  bx_guard.iaddr.lin[bx_guard.iaddr.num_linear].bpoint_id = bx_debugger.next_bpoint_id++;
  bx_guard.iaddr.num_linear++;
  bx_guard.guard_for |= BX_DBG_GUARD_IADDR_LIN;

#else
  fprintf(stderr, "Error: linear breakpoint support not compiled in.\n");
  fprintf(stderr, "Error: see BX_DBG_SUPPORT_LIN_BPOINT.\n");
#endif
}

  void
bx_dbg_pbreakpoint_command(Boolean specific, Bit32u paddress)
{
#if BX_DBG_SUPPORT_PHY_BPOINT
  if (specific == 0) {
    fprintf(stderr, "Error: pbreak without address not implemented yet.\n");
    return;
    }

  if (bx_guard.iaddr.num_physical >= BX_DBG_MAX_PHY_BPOINTS) {
    fprintf(stderr, "Error: no more physical breakpoint slots left.\n");
    fprintf(stderr, "Error: see BX_DBG_MAX_PHY_BPOINTS.\n");
    return;
    }

  bx_guard.iaddr.phy[bx_guard.iaddr.num_physical].addr = paddress;
  bx_guard.iaddr.phy[bx_guard.iaddr.num_physical].bpoint_id = bx_debugger.next_bpoint_id++;
  bx_guard.iaddr.num_physical++;
  bx_guard.guard_for |= BX_DBG_GUARD_IADDR_PHY;

#else
  fprintf(stderr, "Error: physical breakpoint support not compiled in.\n");
  fprintf(stderr, "Error: see BX_DBG_SUPPORT_PHY_BPOINT.\n");
#endif
}

  void
bx_dbg_info_bpoints_command(void)
{
  unsigned i;
// Num Type           Disp Enb Address    What
// 1   breakpoint     keep y   0x00010664 in main at temp.c:7

  fprintf(stderr, "Num Type           Disp Enb Address\n");
#if BX_DBG_SUPPORT_VIR_BPOINT
  for (i=0; i<bx_guard.iaddr.num_virtual; i++) {
    fprintf(stderr, "%3u ", bx_guard.iaddr.vir[i].bpoint_id);
    fprintf(stderr, "vbreakpoint    ");
    fprintf(stderr, "keep ");
    fprintf(stderr, "y   ");
    fprintf(stderr, "0x%04x:0x%08x\n",
                  bx_guard.iaddr.vir[i].cs,
                  bx_guard.iaddr.vir[i].eip);
    }
#endif

#if BX_DBG_SUPPORT_LIN_BPOINT
  for (i=0; i<bx_guard.iaddr.num_linear; i++) {
    fprintf(stderr, "%3u ", bx_guard.iaddr.lin[i].bpoint_id);
    fprintf(stderr, "lbreakpoint    ");
    fprintf(stderr, "keep ");
    fprintf(stderr, "y   ");
    fprintf(stderr, "0x%08x\n",
                  bx_guard.iaddr.lin[i].addr);
    }
#endif

#if BX_DBG_SUPPORT_PHY_BPOINT
  for (i=0; i<bx_guard.iaddr.num_physical; i++) {
    fprintf(stderr, "%3u ", bx_guard.iaddr.phy[i].bpoint_id);
    fprintf(stderr, "pbreakpoint    ");
    fprintf(stderr, "keep ");
    fprintf(stderr, "y   ");
    fprintf(stderr, "0x%08x\n",
                  bx_guard.iaddr.phy[i].addr);
    }
#endif
}


  void
bx_dbg_set_command(char *p1, char *p2, char *p3)
{
  fprintf(stderr, "Error: %s %s %s: command 'set' not implemented yet.\n",
    p1, p2, p3);
}

  void
bx_dbg_take_command(char *what, unsigned n)
{
  if ( !strcmp(what, "dma") ) {
    unsigned i;
    if (n == 0) {
      fprintf(stderr, "Error: take what n=0.\n");
      return;
      }
    bx_dbg_post_dma_reports(); // in case there's some pending reports
    bx_dbg_batch_dma.this_many = n;

    for (i=0; i<n; i++) {
      bx_dbg_callback[0].take_dma();
      }

    bx_dbg_batch_dma.this_many = 1;  // reset to normal
    bx_dbg_post_dma_reports(); // print reports and flush
    if (bx_guard.report.dma)
      fprintf(stderr, "done\n");
    }
  else if ( !strcmp(what, "irq") ) {
    bx_dbg_callback[0].take_irq();

    if (bx_guard.report.irq)
      fprintf(stderr, "done\n");
    }
  else {
    fprintf(stderr, "Error: Take '%s' not understood.\n", what);
    }
}


  void
bx_dbg_info_registers_command(void)
{
  Bit32u reg;
  bx_dbg_cpu_t cpu;

  for (unsigned i=0; i<BX_NUM_SIMULATORS; i++) {
    memset(&cpu, 0, sizeof(cpu));
    bx_dbg_callback[i].get_cpu(&cpu);

#if (BX_NUM_SIMULATORS > 1)
    fprintf(stderr, "SIM[%u]:\n", i);
#endif
    reg = cpu.eax;
    fprintf(stderr, "eax            0x%-8x\t%d\n", (unsigned) reg, (int) reg);
    reg = cpu.ecx;
    fprintf(stderr, "ecx            0x%-8x\t%d\n", (unsigned) reg, (int) reg);
    reg = cpu.edx;
    fprintf(stderr, "edx            0x%-8x\t%d\n", (unsigned) reg, (int) reg);
    reg = cpu.ebx;
    fprintf(stderr, "ebx            0x%-8x\t%d\n", (unsigned) reg, (int) reg);

    reg = cpu.esp;
    fprintf(stderr, "esp            0x%-8x\t0x%-8x\n", (unsigned) reg, (int) reg);
    reg = cpu.ebp;
    fprintf(stderr, "ebp            0x%-8x\t0x%-8x\n", (unsigned) reg, (int) reg);
    reg = cpu.esi;
    fprintf(stderr, "esi            0x%-8x\t%d\n", (unsigned) reg, (int) reg);
    reg = cpu.edi;
    fprintf(stderr, "edi            0x%-8x\t%d\n", (unsigned) reg, (int) reg);

    reg = cpu.eip;
    fprintf(stderr, "eip            0x%-8x\t0x%-8x\n", (unsigned) reg, (int) reg);

    reg = cpu.eflags;
    fprintf(stderr, "eflags         0x%-8x\t%d\n", (unsigned) reg, (int) reg);

    reg = cpu.cs.sel;
    fprintf(stderr, "cs             0x%-8x\t%d\n", (unsigned) reg, (int) reg);
    reg = cpu.ss.sel;
    fprintf(stderr, "ss             0x%-8x\t%d\n", (unsigned) reg, (int) reg);
    reg = cpu.ds.sel;
    fprintf(stderr, "ds             0x%-8x\t%d\n", (unsigned) reg, (int) reg);
    reg = cpu.es.sel;
    fprintf(stderr, "es             0x%-8x\t%d\n", (unsigned) reg, (int) reg);
    reg = cpu.fs.sel;
    fprintf(stderr, "fs             0x%-8x\t%d\n", (unsigned) reg, (int) reg);
    reg = cpu.gs.sel;
    fprintf(stderr, "gs             0x%-8x\t%d\n", (unsigned) reg, (int) reg);
    }
}

  void
bx_dbg_info_program_command(void)
{
  fprintf(stderr, "        Using the running image of child process -1.\n");
  fprintf(stderr, "Program stopped at 0x0.\n");
  fprintf(stderr, "It stopped at breakpoint 0.\n");
}


  void
bx_dbg_dump_cpu_command(void)
{
  bx_dbg_cpu_t cpu;

  for (unsigned i=0; i<BX_NUM_SIMULATORS; i++ ) {
    bx_dbg_callback[i].get_cpu(&cpu);

#if (BX_NUM_SIMULATORS >= 2)
    fprintf(stderr, "CPU#%u\n", i);
#endif
    fprintf(stderr, "eax:0x%x\n", (unsigned) cpu.eax);
    fprintf(stderr, "ebx:0x%x\n", (unsigned) cpu.ebx);
    fprintf(stderr, "ecx:0x%x\n", (unsigned) cpu.ecx);
    fprintf(stderr, "edx:0x%x\n", (unsigned) cpu.edx);

    fprintf(stderr, "ebp:0x%x\n", (unsigned) cpu.ebp);
    fprintf(stderr, "esi:0x%x\n", (unsigned) cpu.esi);
    fprintf(stderr, "edi:0x%x\n", (unsigned) cpu.edi);
    fprintf(stderr, "esp:0x%x\n", (unsigned) cpu.esp);

    fprintf(stderr, "eflags:0x%x\n", (unsigned) cpu.eflags);
    fprintf(stderr, "eip:0x%x\n", (unsigned) cpu.eip);

    fprintf(stderr, "cs:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
      (unsigned) cpu.cs.sel, (unsigned) cpu.cs.des_l,
      (unsigned) cpu.cs.des_h, (unsigned) cpu.cs.valid);

    fprintf(stderr, "ss:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
      (unsigned) cpu.ss.sel, (unsigned) cpu.ss.des_l,
      (unsigned) cpu.ss.des_h, (unsigned) cpu.ss.valid);

    fprintf(stderr, "ds:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
      (unsigned) cpu.ds.sel, (unsigned) cpu.ds.des_l,
      (unsigned) cpu.ds.des_h, (unsigned) cpu.ds.valid);

    fprintf(stderr, "es:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
      (unsigned) cpu.es.sel, (unsigned) cpu.es.des_l,
      (unsigned) cpu.es.des_h, (unsigned) cpu.es.valid);

    fprintf(stderr, "fs:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
      (unsigned) cpu.fs.sel, (unsigned) cpu.fs.des_l,
      (unsigned) cpu.fs.des_h, (unsigned) cpu.fs.valid);

    fprintf(stderr, "gs:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
      (unsigned) cpu.gs.sel, (unsigned) cpu.gs.des_l,
      (unsigned) cpu.gs.des_h, (unsigned) cpu.gs.valid);

    fprintf(stderr, "ldtr:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
      (unsigned) cpu.ldtr.sel, (unsigned) cpu.ldtr.des_l,
      (unsigned) cpu.ldtr.des_h, (unsigned) cpu.ldtr.valid);

    fprintf(stderr, "tr:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
      (unsigned) cpu.tr.sel, (unsigned) cpu.tr.des_l,
      (unsigned) cpu.tr.des_h, (unsigned) cpu.tr.valid);

    fprintf(stderr, "gdtr:base=0x%x, limit=0x%x\n",
      (unsigned) cpu.gdtr.base, (unsigned) cpu.gdtr.limit);

    fprintf(stderr, "idtr:base=0x%x, limit=0x%x\n",
      (unsigned) cpu.idtr.base, (unsigned) cpu.idtr.limit);

    fprintf(stderr, "dr0:0x%x\n", (unsigned) cpu.dr0);
    fprintf(stderr, "dr1:0x%x\n", (unsigned) cpu.dr1);
    fprintf(stderr, "dr2:0x%x\n", (unsigned) cpu.dr2);
    fprintf(stderr, "dr3:0x%x\n", (unsigned) cpu.dr3);
    fprintf(stderr, "dr4:0x%x\n", (unsigned) cpu.dr4);
    fprintf(stderr, "dr5:0x%x\n", (unsigned) cpu.dr5);
    fprintf(stderr, "dr6:0x%x\n", (unsigned) cpu.dr6);
    fprintf(stderr, "dr7:0x%x\n", (unsigned) cpu.dr7);

    fprintf(stderr, "tr3:0x%x\n", (unsigned) cpu.tr3);
    fprintf(stderr, "tr4:0x%x\n", (unsigned) cpu.tr4);
    fprintf(stderr, "tr5:0x%x\n", (unsigned) cpu.tr5);
    fprintf(stderr, "tr6:0x%x\n", (unsigned) cpu.tr6);
    fprintf(stderr, "tr7:0x%x\n", (unsigned) cpu.tr7);

    fprintf(stderr, "cr0:0x%x\n", (unsigned) cpu.cr0);
    fprintf(stderr, "cr1:0x%x\n", (unsigned) cpu.cr1);
    fprintf(stderr, "cr2:0x%x\n", (unsigned) cpu.cr2);
    fprintf(stderr, "cr3:0x%x\n", (unsigned) cpu.cr3);
    fprintf(stderr, "cr4:0x%x\n", (unsigned) cpu.cr4);

    fprintf(stderr, "inhibit_int:%u\n", cpu.inhibit_int);
    }

#if BX_PCI_SUPPORT
  if (bx_options.i440FXSupport) {
    bx_devices.pci->print_i440fx_state(stderr);
    }
#endif

  fprintf(stderr, "done\n");
}


  void
bx_dbg_examine_command(char *command, char *format, Boolean format_passed,
               Bit32u addr, Boolean addr_passed)
{
  unsigned repeat_count, i;
  char ch, display_format, unit_size;
  Boolean iteration;
  unsigned data_size;
  Boolean paddr_valid;
  Bit32u  paddr;
  Bit8u   data8;
  Bit16u  data16;
  Bit32u  data32;
  unsigned columns, per_line, offset;
  unsigned char digit;
  unsigned biti;
  Boolean is_linear;
  unsigned char databuf[8];

  // If command was the extended "xp" command, meaning eXamine Physical memory,
  // then flag memory address as physical, rather than linear.
  if (strcmp(command, "xp") == 0) {
    is_linear = 0;
    }
  else {
    is_linear = 1;
    }

  if (addr_passed==0)
    addr = bx_debugger.default_addr;

  if (format_passed==0) {
    display_format = bx_debugger.default_display_format;
    unit_size      = bx_debugger.default_unit_size;
    repeat_count   = 1;
    }
  else {
    if (format==NULL) {
      fprintf(stderr, "dbg_examine: format NULL\n");
      bx_dbg_exit(1);
      }

    if (strlen(format) < 2) {
      fprintf(stderr, "dbg_examine: invalid format passed.\n");
      bx_dbg_exit(1);
      }

    if (format[0] != '/') {
      fprintf(stderr, "dbg_examine: '/' is not first char of format.\n");
      bx_dbg_exit(1);
      }

    format++;
    repeat_count = 0;
    ch = *format;
    iteration = 0;

    while ( (ch>='0') && (ch<='9') ) {
      iteration = 1;
      repeat_count = 10*repeat_count + (ch-'0');
      format++;
      ch = *format;
      }

    if (iteration==0) {
      // if no count given, use default
      repeat_count = 1;
      }
    else if (repeat_count==0) {
      // count give, but zero is an error
      fprintf(stderr, "dbg_examine: repeat count given but is zero.\n");
      return;
      }


    // set up the default display format and unit size parameters
    display_format = bx_debugger.default_display_format;
    unit_size      = bx_debugger.default_unit_size;

    for (i=0; i<=1; i++) {
      if (ch==0) break; // bail on null character
      switch (ch) {
        case 'x': // hex
        case 'd': // signed decimal
        case 'u': // unsigned decimal
        case 'o': // octal
        case 't': // binary
        case 's': // null terminated string
        case 'i': // machine instruction
          display_format = ch;
          break;

        case 'b': // bytes
        case 'h': // halfwords (two bytes)
        case 'w': // words (4 bytes)
        case 'g': // giant words (8 bytes)
          unit_size = ch;
          break;
        default:
          fprintf(stderr, "dbg_examine: invalid format passed.\n");
          bx_dbg_exit(1);
          break;
        }
      format++;
      ch = *format;
      }

    // store current options as default
    bx_debugger.default_display_format = display_format;
    bx_debugger.default_unit_size      = unit_size;
    }

  //fprintf(stderr, "  repeat count was %u\n", repeat_count);
  //fprintf(stderr, "  display_format = '%c'\n", display_format);
  //fprintf(stderr, "  unit_size      = '%c'\n", unit_size);

  if ( (display_format == 'i') || (display_format == 's') ) {
    fprintf(stderr, "error: dbg_examine: 'i' and 's' formats not supported.\n");
    return;
    }

  if (unit_size == 'g') {
    fprintf(stderr, "error: dbg_examine: 'g' (8-byte) unit size not supported.\n");
    return;
    }

  data_size = 0;
  per_line  = 0;
  offset = 0;

  switch (unit_size) {
    case 'b': data_size = 1; per_line = 8; break;
    case 'h': data_size = 2; per_line = 8; break;
    case 'w': data_size = 4; per_line = 4; break;
    //case 'g': data_size = 8; per_line = 2; break;
    }

  columns = per_line + 1; // set current number columns past limit

  for (i=1; i<=repeat_count; i++) {

    if (columns > per_line) {
      // if not 1st run, need a newline from last line
      if (i!=1)
        fprintf(stderr, "\n");
      fprintf(stderr, "0x%x <bogus+%u>:", addr, offset);
      columns = 1;
      }

    if (is_linear) {
      bx_dbg_callback[0].xlate_linear2phy(addr, &paddr, &paddr_valid);
      if (!paddr_valid) {
        fprintf(stderr, "error: examine memory: no tranlation for linear-to-phy mem available.\n");
        return;
        }
      }
    else {
      paddr = addr;  // address is already physical address
      }

    bx_dbg_callback[0].getphymem(paddr, data_size, databuf);
    switch (data_size) {
      case 1:
        data8 = databuf[0];
        switch (display_format) {
          case 'x': fprintf(stderr, "\t0x%02x", (unsigned) data8); break;
          case 'd': fprintf(stderr, "\t%d", (int) (Bit8s) data8); break;
          case 'u': fprintf(stderr, "\t%u", (unsigned) data8); break;
          case 'o': fprintf(stderr, "\t%o", (unsigned) data8); break;
          case 't':
            fputc('\t', stderr);
            for (biti=7; ; biti--) {
              digit = (data8 >> biti) & 0x01;
              fputc(digit + '0', stderr);
              if (biti==0) break;
              }
            break;
          }
        break;

      case 2:
#ifdef BX_LITTLE_ENDIAN
        data16 = * (Bit16u *) databuf;
#else
        data16 = (databuf[1]<<8)  |  databuf[0];
#endif
        switch (display_format) {
          case 'x': fprintf(stderr, "\t0x%04x", (unsigned) data16); break;
          case 'd': fprintf(stderr, "\t%d", (int) (Bit16s) data16); break;
          case 'u': fprintf(stderr, "\t%u", (unsigned) data16); break;
          case 'o': fprintf(stderr, "\t%o", (unsigned) data16); break;
          case 't':
            fputc('\t', stderr);
            for (biti=15; ; biti--) {
              digit = (data16 >> biti) & 0x01;
              fputc(digit + '0', stderr);
              if (biti==0) break;
              }
            break;
          }
        break;

      case 4:
#ifdef BX_LITTLE_ENDIAN
        data32 = * (Bit32u *) databuf;
#else
        data32 = (databuf[3]<<24) | (databuf[2]<<16) |
                 (databuf[1]<<8)  |  databuf[0];
#endif
        switch (display_format) {
          case 'x': fprintf(stderr, "\t0x%08x", (unsigned) data32); break;
          case 'd': fprintf(stderr, "\t%d", (int) (Bit32s) data32); break;
          case 'u': fprintf(stderr, "\t%u", (unsigned) data32); break;
          case 'o': fprintf(stderr, "\t%o", (unsigned) data32); break;
          case 't':
            fputc('\t', stderr);
            for (biti=31; ; biti--) {
              digit = (data32 >> biti) & 0x01;
              fputc(digit + '0', stderr);
              if (biti==0) break;
              }
            break;
          }
        break;
      }

    addr += data_size;
    bx_debugger.default_addr = addr;
    columns++;
    offset += data_size;
    }
  fprintf(stderr, "\n");
}

  void
bx_dbg_setpmem_command(Bit32u addr, unsigned len, Bit32u val)
{
  Boolean is_OK;
  Bit8u   buf[4];

  switch ( len ) {
    case 1:
      buf[0] = (Bit8u) val;
      break;
    case 2:
      buf[0] = val & 0xff;
      buf[1] = (val>>8) & 0xff;
      break;
    case 4:
      buf[0] = val & 0xff; val >>= 8;
      buf[1] = val & 0xff; val >>= 8;
      buf[2] = val & 0xff; val >>= 8;
      buf[3] = val & 0xff;
      break;
    default:
      fprintf(stderr, "Error: setpmem: bad length value = %u\n", len);
      return;
    }

  is_OK = bx_dbg_callback[0].setphymem(addr, len, buf);
  if (!is_OK) {
    fprintf(stderr, "Error: setpmem: could not set memory, out of physical bounds?\n");
    }
}

  void
bx_dbg_set_symbol_command(char *symbol, Bit32u val)
{
  Boolean is_OK;
  symbol++; // get past '$'

  if ( !strcmp(symbol, "eax") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_EAX, val);
    }
  else if ( !strcmp(symbol, "ecx") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_ECX, val);
    }
  else if ( !strcmp(symbol, "edx") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_EDX, val);
    }
  else if ( !strcmp(symbol, "ebx") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_EBX, val);
    }
  else if ( !strcmp(symbol, "esp") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_ESP, val);
    }
  else if ( !strcmp(symbol, "ebp") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_EBP, val);
    }
  else if ( !strcmp(symbol, "esi") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_ESI, val);
    }
  else if ( !strcmp(symbol, "edi") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_EDI, val);
    }
  else if ( !strcmp(symbol, "eip") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_EIP, val);
    }
  else if ( !strcmp(symbol, "eflags") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_EFLAGS, val);
    }
  else if ( !strcmp(symbol, "cs") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_CS, val);
    }
  else if ( !strcmp(symbol, "ss") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_SS, val);
    }
  else if ( !strcmp(symbol, "ds") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_DS, val);
    }
  else if ( !strcmp(symbol, "es") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_ES, val);
    }
  else if ( !strcmp(symbol, "fs") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_FS, val);
    }
  else if ( !strcmp(symbol, "gs") ) {
    is_OK = bx_dbg_callback[0].set_reg(BX_DBG_REG_GS, val);
    }
  else if ( !strcmp(symbol, "synchronous_dma") ) {
    bx_guard.async.dma = !val;
    return;
    }
  else if ( !strcmp(symbol, "synchronous_irq") ) {
    bx_guard.async.irq = !val;
    return;
    }
  else if ( !strcmp(symbol, "event_reports") ) {
    bx_guard.report.irq   = val;
    bx_guard.report.a20   = val;
    bx_guard.report.io    = val;
    bx_guard.report.ucmem = val;
    bx_guard.report.dma   = val;
    return;
    }
  else if ( !strcmp(symbol, "auto_disassemble") ) {
    bx_debugger.auto_disassemble = (val > 0);
    return;
    }
  else if ( !strcmp(symbol, "disassemble_size") ) {
    if ( (val!=16) && (val!=32) ) {
      fprintf(stderr, "Error: disassemble_size must be 16 or 32.\n");
      return;
      }
    bx_debugger.disassemble_size = val;
    return;
    }
  else {
    fprintf(stderr, "Error: set: unrecognized symbol.\n");
    return;
    }

  if (!is_OK) {
    fprintf(stderr, "Error: could not set register '%s'.\n", symbol);
    }
}

  void
bx_dbg_query_command(char *what)
{
  unsigned pending;

  if ( !strcmp(what, "pending") ) {
    pending = bx_dbg_callback[0].query_pending();

    if ( pending & BX_DBG_PENDING_DMA )
      fprintf(stderr, "pending DMA\n");

    if ( pending & BX_DBG_PENDING_IRQ )
      fprintf(stderr, "pending IRQ\n");

    if (!pending)
      fprintf(stderr, "pending none\n");

    fprintf(stderr, "done\n");
    }
  else {
    fprintf(stderr, "Error: Query '%s' not understood.\n", what);
    }
}

  void
bx_dbg_set_cpu_command(void)
{
  FILE *fp;
  int   reti;
  char *rets;
  Boolean retb;
  unsigned long ul1, ul2, ul3, ul4;

  bx_dbg_cpu_t cpu;

  fp = bx_infile_stack[bx_infile_stack_index].fp;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "eax:0x%lx", &ul1); cpu.eax = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "ebx:0x%lx", &ul1); cpu.ebx = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "ecx:0x%lx", &ul1); cpu.ecx = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "edx:0x%lx", &ul1); cpu.edx = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "ebp:0x%lx", &ul1); cpu.ebp = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "esi:0x%lx", &ul1); cpu.esi = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "edi:0x%lx", &ul1); cpu.edi = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "esp:0x%lx", &ul1); cpu.esp = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "eflags:0x%lx", &ul1); cpu.eflags = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "eip:0x%lx", &ul1); cpu.eip = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "cs:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &ul1, &ul2, &ul3, &ul4);
  cpu.cs.sel   = (Bit16u) ul1;
  cpu.cs.des_l = ul2;
  cpu.cs.des_h = ul3;
  cpu.cs.valid = ul4;
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "ss:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &ul1, &ul2, &ul3, &ul4);
  cpu.ss.sel   = (Bit16u) ul1;
  cpu.ss.des_l = ul2;
  cpu.ss.des_h = ul3;
  cpu.ss.valid = ul4;
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "ds:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &ul1, &ul2, &ul3, &ul4);
  cpu.ds.sel   = (Bit16u) ul1;
  cpu.ds.des_l = ul2;
  cpu.ds.des_h = ul3;
  cpu.ds.valid = ul4;
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "es:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &ul1, &ul2, &ul3, &ul4);
  cpu.es.sel   = (Bit16u) ul1;
  cpu.es.des_l = ul2;
  cpu.es.des_h = ul3;
  cpu.es.valid = ul4;
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "fs:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &ul1, &ul2, &ul3, &ul4);
  cpu.fs.sel   = (Bit16u) ul1;
  cpu.fs.des_l = ul2;
  cpu.fs.des_h = ul3;
  cpu.fs.valid = ul4;
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "gs:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &ul1, &ul2, &ul3, &ul4);
  cpu.gs.sel   = (Bit16u) ul1;
  cpu.gs.des_l = ul2;
  cpu.gs.des_h = ul3;
  cpu.gs.valid = ul4;
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "ldtr:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &ul1, &ul2, &ul3, &ul4);
  cpu.ldtr.sel   = (Bit16u) ul1;
  cpu.ldtr.des_l = ul2;
  cpu.ldtr.des_h = ul3;
  cpu.ldtr.valid = ul4;
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "tr:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &ul1, &ul2, &ul3, &ul4);
  cpu.tr.sel   = (Bit16u) ul1;
  cpu.tr.des_l = ul2;
  cpu.tr.des_h = ul3;
  cpu.tr.valid = ul4;
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "gdtr:base=0x%lx, limit=0x%lx",
                &ul1, &ul2);
  cpu.gdtr.base  = ul1;
  cpu.gdtr.limit = ul2;
  if (reti != 2) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "idtr:base=0x%lx, limit=0x%lx",
                &ul1, &ul2);
  cpu.idtr.base  = ul1;
  cpu.idtr.limit = ul2;
  if (reti != 2) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr0:0x%lx", &ul1); cpu.dr0 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr1:0x%lx", &ul1); cpu.dr1 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr2:0x%lx", &ul1); cpu.dr2 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr3:0x%lx", &ul1); cpu.dr3 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr4:0x%lx", &ul1); cpu.dr4 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr5:0x%lx", &ul1); cpu.dr5 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr6:0x%lx", &ul1); cpu.dr6 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr7:0x%lx", &ul1); cpu.dr7 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "tr3:0x%lx", &ul1); cpu.tr3 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "tr4:0x%lx", &ul1); cpu.tr4 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "tr5:0x%lx", &ul1); cpu.tr5 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "tr6:0x%lx", &ul1); cpu.tr6 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "tr7:0x%lx", &ul1); cpu.tr7 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "cr0:0x%lx", &ul1); cpu.cr0 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "cr1:0x%lx", &ul1); cpu.cr1 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "cr2:0x%lx", &ul1); cpu.cr2 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "cr3:0x%lx", &ul1); cpu.cr3 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "cr4:0x%lx", &ul1); cpu.cr4 = ul1;
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "inhibit_int:%u", &cpu.inhibit_int);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "done");
  if (reti != 0) goto scanf_error;

  retb = bx_dbg_callback[0].set_cpu(&cpu);
  if (retb == 0)
    fprintf(stderr, "Error: set_cpu encountered error\n");
  else
    fprintf(stderr, "OK\n");
  return;

eof_error:
  fprintf(stderr, "Error: EOF encountered in set_cpu input stream\n");
  return;

scanf_error:
  fprintf(stderr, "Error: scanf returned error in set_cpu input stream\n");
  return;
}

  void
bx_dbg_disassemble_command(Bit32u addr1, Bit32u addr2)
{
#if BX_DISASM
  Boolean paddr_valid;
  Bit32u  paddr;
  unsigned ilen;


  do {
    bx_dbg_callback[0].xlate_linear2phy(addr1, &paddr, &paddr_valid);

    if (paddr_valid) {
      bx_dbg_callback[0].getphymem(paddr, 16, bx_disasm_ibuf);
      ilen = bx_disassemble.disasm(bx_debugger.disassemble_size==32,
                                   bx_disasm_ibuf, bx_disasm_tbuf);

      fprintf(stderr, "%08x: ", (unsigned) addr1);
      for (unsigned j=0; j<ilen; j++)
        fprintf(stderr, "%02x", (unsigned) bx_disasm_ibuf[j]);
      fprintf(stderr, ": %s\n", bx_disasm_tbuf);
      }
    else {
      fprintf(stderr, "??? (physical address not available)\n");
      ilen = 0; // keep compiler happy
      addr1 = addr2; // bail out
      }
      addr1 += ilen;
    } while (addr1 < addr2);
#else
  UNUSED(addr1);
  UNUSED(addr2);
#endif  // #if BX_DISASM
}

  void
bx_dbg_instrument_command(char *comm)
{
#if BX_INSTRUMENTATION
  if ( !strcmp(comm, "start") ) {
    bx_dbg_callback[0].instr_start();
    }
  else if ( !strcmp(comm, "stop") ) {
    bx_dbg_callback[0].instr_stop();
    }
  else if ( !strcmp(comm, "reset") ) {
    bx_dbg_callback[0].instr_reset();
    }
  else if ( !strcmp(comm, "print") ) {
    bx_dbg_callback[0].instr_print();
    }
  else {
    fprintf(stderr, "Error: command instrument %s not implemented.\n", comm);
    bx_dbg_exit(1);
    }
#else
    UNUSED(comm);

    fprintf(stderr, "Error: instrumentation not enabled.\n");
#endif
}

  void
bx_dbg_loader_command(char *path_quoted)
{
  size_t len;

  // skip beginning double quote
  if (path_quoted[0] == '"')
    path_quoted++;

  // null out ending quote
  len = strlen(path_quoted);
  if (path_quoted[len - 1] == '"')
    path_quoted[len - 1] = '\0';

#if BX_USE_LOADER
  bx_dbg_callback[0].loader(path_quoted);
#else
  fprintf(stderr, "Error: loader not implemented.\n");
#endif
}

  void
bx_dbg_doit_command(unsigned n)
{
  // generic command to add temporary hacks to
  // for debugging purposes
  UNUSED(n);

#if 0
void init_collect_data(void);
void show_collect_data(void);

  if (n==0)
    init_collect_data();
  else
    show_collect_data();
#endif

#if 0
doit = 1;
  fprintf(stderr, "IO journal size = %u\n", bx_debugger.IO_journal.size);
  for (unsigned i=0; i<bx_debugger.IO_journal.size; i++) {
    fprintf(stderr, "[%u].op = %u\n", i,
      bx_debugger.IO_journal.element[i].op);
    fprintf(stderr, "[%u].len = %u\n", i,
      bx_debugger.IO_journal.element[i].len);
    fprintf(stderr, "[%u].addr = %u\n", i,
      bx_debugger.IO_journal.element[i].addr);
    fprintf(stderr, "[%u].value = %u\n", i,
      bx_debugger.IO_journal.element[i].value);
    }
#endif

  bx_dbg.interrupts = n;
  bx_dbg.exceptions = n;
}

  void
bx_dbg_crc_command(Bit32u addr1, Bit32u addr2)
{
  Bit32u crc1, crc2;

  if (addr1 >= addr2) {
    fprintf(stderr, "Error: crc: invalid range.\n");
    return;
    }

  if (!bx_dbg_callback[0].crc32(crc32, addr1, addr2, &crc1)) {
    fprintf(stderr, "sim0: could not CRC memory\n");
    return;
    }
#if BX_NUM_SIMULATORS == 1
  fprintf(stderr, "0x%lx\n", crc1);
#else
  if (!bx_dbg_callback[1].crc32(crc32, addr1, addr2, &crc2)) {
    fprintf(stderr, "sim1: could not CRC memory\n");
    return;
    }
  if (crc1 == crc2) {
    fprintf(stderr, "CRC same: 0x%x\n", (unsigned) crc1);
    }
  else {
    fprintf(stderr, "CRC different: sim0=0x%x, sim1=0x%x\n",
            (unsigned) crc1, (unsigned) crc2);
    }
#endif
}

  void
bx_dbg_info_dirty_command(void)
{
  unsigned char *page_tbl = bx_dbg_callback[0].dirty_page_tbl;
  unsigned page_tbl_size  = bx_dbg_callback[0].dirty_page_tbl_size;

  for (unsigned i=0; i<page_tbl_size; i++) {
    if (page_tbl[i]) {
      fprintf(stderr, "0x%x\n", i);
      page_tbl[i] = 0; // reset to clean
      }
    }
}





//
// Reports from various events
//

  void
bx_dbg_iac_report(unsigned vector, unsigned irq)
{
#if BX_NUM_SIMULATORS > 1
  unsigned tail, master;
#endif

if (doit) fprintf(stderr, "iac report: vector=%u\n", vector);

  if (bx_guard.report.irq) {
    fprintf(stderr, "event icount=%u IRQ irq=%u vec=%x\n",
      (unsigned) bx_guard_found[0].icount, irq, vector);
    }

#if BX_NUM_SIMULATORS > 1
  if (bx_debugger.master_slave_mode == BX_DBG_SLAVE_MODE ) {
    fprintf(stderr, "Error: iac_report: in slave mode.\n");
    bx_dbg_exit(1);
    }

  // Master simulator mode
  if (bx_debugger.async_journal.size >= BX_DBG_ASYNC_JOURNAL_SIZE) {
    fprintf(stderr, "Error: iac: async journal full.\n");
    bx_dbg_exit(1);
    }

  if (bx_debugger.async_journal.size == 0) {
    // start off point head & tail at same element
    bx_debugger.async_journal.head = 0;
    tail = bx_debugger.async_journal.tail = 0;
    }
  else {
    tail = bx_debugger.async_journal.tail + 1;
    }
  if (tail >= BX_DBG_ASYNC_JOURNAL_SIZE) {
    fprintf(stderr, "Error: iac_report: journal wrapped.\n");
    bx_dbg_exit(0);
    }

  master = bx_debugger.master;
  bx_debugger.async_journal.element[tail].what = BX_DBG_ASYNC_JOURNAL_IAC;
  bx_debugger.async_journal.element[tail].icount = bx_guard_found[master].icount;
  bx_debugger.async_journal.element[tail].u.iac.val = vector;

  if (bx_debugger.async_journal.size)
    bx_debugger.async_journal.tail++;
  bx_debugger.async_journal.size++;
#endif
}

  void
bx_dbg_a20_report(unsigned val)
{
  if (bx_guard.report.a20) {
    fprintf(stderr, "event icount=%u A20 val=%u\n",
      (unsigned) bx_guard_found[0].icount, val);
    }
}

#if BX_NUM_SIMULATORS > 1
  void
bx_dbg_journal_a20_event(unsigned val)
{
  unsigned tail, master;

  if (bx_debugger.master_slave_mode == BX_DBG_SLAVE_MODE ) {
    fprintf(stderr, "Error: a20_report: in slave mode.\n");
    bx_dbg_exit(1);
    }

  // Master simulator mode
  if (bx_debugger.async_journal.size >= BX_DBG_ASYNC_JOURNAL_SIZE) {
    fprintf(stderr, "Error: async journal full.\n");
    bx_dbg_exit(1);
    }

  if (bx_debugger.async_journal.size == 0) {
    // start off point head & tail at same element
    bx_debugger.async_journal.head = 0;
    tail = bx_debugger.async_journal.tail = 0;
    }
  else {
    tail = bx_debugger.async_journal.tail + 1;
    }
  if (tail >= BX_DBG_ASYNC_JOURNAL_SIZE) {
    fprintf(stderr, "Error: a20_report: journal wrapped.\n");
    bx_dbg_exit(0);
    }

  master = bx_debugger.master;
  bx_debugger.async_journal.element[tail].what = BX_DBG_ASYNC_JOURNAL_A20;
  bx_debugger.async_journal.element[tail].icount = bx_guard_found[master].icount;
  bx_debugger.async_journal.element[tail].u.a20.val = val;

  if (bx_debugger.async_journal.size)
    bx_debugger.async_journal.tail++;
  bx_debugger.async_journal.size++;
}
#endif

  void
bx_dbg_io_report(Bit32u addr, unsigned size, unsigned op, Bit32u val)
{
  if (bx_guard.report.io) {
    fprintf(stderr, "event icount=%u IO addr=0x%x size=%u op=%s val=0x%x\n",
      (unsigned) bx_guard_found[0].icount,
      (unsigned) addr,
      size,
      (op==BX_READ) ? "read" : "write",
      (unsigned) val);
    }

  // nothing else to do.  bx_dbg_inp() and bx_dbg_outp() do the journaling.
}

  void
bx_dbg_ucmem_report(Bit32u addr, unsigned size, unsigned op, Bit32u val)
{
  if (bx_guard.report.ucmem) {
    fprintf(stderr, "event icount=%u UCmem addr=0x%x size=%u op=%s val=0x%x\n",
      (unsigned) bx_guard_found[0].icount,
      (unsigned) addr,
      size,
      (op==BX_READ) ? "read" : "write",
      (unsigned) val);
    }
  // nothing else to do.  bx_dbg_ucmem_read() and bx_dbg_ucmem_write()
  // do the journaling.
}

  void
bx_dbg_dma_report(Bit32u addr, unsigned len, unsigned what, Bit32u val)
{
  if (bx_dbg_batch_dma.this_many == 0) {
    fprintf(stderr, "%s: DMA batch this_many=0.\n", argv0);
    bx_dbg_exit(1);
    }

  // if Q is full, post events (and flush)
  if (bx_dbg_batch_dma.Qsize >= bx_dbg_batch_dma.this_many) {
    fprintf(stderr, "%s: DMA batch Q was not flushed.\n", argv0);
    bx_dbg_exit(1);
    }

  // if Q already has MAX elements in it
  if (bx_dbg_batch_dma.Qsize >= BX_BATCH_DMA_BUFSIZE) {
    fprintf(stderr, "%s: DMA batch buffer overrun.\n", argv0);
    bx_dbg_exit(1);
    }

  bx_dbg_batch_dma.Qsize++;
  bx_dbg_batch_dma.Q[bx_dbg_batch_dma.Qsize-1].addr   = addr;
  bx_dbg_batch_dma.Q[bx_dbg_batch_dma.Qsize-1].len    = len;
  bx_dbg_batch_dma.Q[bx_dbg_batch_dma.Qsize-1].what   = what;
  bx_dbg_batch_dma.Q[bx_dbg_batch_dma.Qsize-1].val    = val;
  bx_dbg_batch_dma.Q[bx_dbg_batch_dma.Qsize-1].icount = bx_guard_found[0].icount;

  // if Q is full, post events (and flush)
  if (bx_dbg_batch_dma.Qsize >= bx_dbg_batch_dma.this_many)
    bx_dbg_post_dma_reports();
}

  void
bx_dbg_post_dma_reports(void)
{
  unsigned i;
  unsigned addr, len, what, val;
  unsigned last_addr, last_len, last_what;
  unsigned print_header;
  unsigned first_iteration;

  if (bx_guard.report.dma) {
    if (bx_dbg_batch_dma.Qsize == 0) return; // nothing batched to print

    // compress output so all contiguous DMA ops of the same type and size
    // are printed on the same line
    last_addr = bx_dbg_batch_dma.Q[0].addr;
    last_len  = bx_dbg_batch_dma.Q[0].len;
    last_what = bx_dbg_batch_dma.Q[0].what;
    first_iteration = 1;

    for (i=0; i<bx_dbg_batch_dma.Qsize; i++) {
      addr = bx_dbg_batch_dma.Q[i].addr;
      len  = bx_dbg_batch_dma.Q[i].len;
      what = bx_dbg_batch_dma.Q[i].what;
      val  = bx_dbg_batch_dma.Q[i].val;

      if (len != last_len)
        print_header = 1;
      else if (what != last_what)
        print_header = 1;
      else if (addr != (last_addr + last_len))
        print_header = 1;
      else
        print_header = 0;

      // now store current values for next iteration
      last_addr = addr;
      last_len  = len;
      last_what = what;

      if (print_header) {
        if (!first_iteration) // need return from previous line
          fprintf(stderr, "\n");
        else
          first_iteration = 0;
        // need to output the event header
        fprintf(stderr, "event icount=%u DMA addr=0x%x size=%u op=%s val=0x%x",
                         (unsigned) bx_dbg_batch_dma.Q[i].icount,
                         addr, len, (what==BX_READ) ? "read" : "write",
                         val );
        print_header = 0;
        }
      else {
        // *no* need to output the event header
        fprintf(stderr, " 0x%x", val);
        }
      }
    if (bx_dbg_batch_dma.Qsize)
      fprintf(stderr, "\n");
    }

  // empty Q, regardless of whether reports are printed
  bx_dbg_batch_dma.Qsize = 0;
}




//
// Cosimulation routines
//

#if (BX_NUM_SIMULATORS >= 2)

  Bit8u
bx_dbg_ucmem_read(Bit32u addr)
{
  Bit8u value;
  unsigned head, tail;

  if ( bx_debugger.master_slave_mode == BX_DBG_MASTER_MODE ) {
    if (bx_debugger.UCmem_journal.size >= BX_DBG_UCMEM_JOURNAL_SIZE) {
      fprintf(stderr, "dbg_ucmem_read: journal full.\n");
      bx_dbg_exit(0);
      }

    if (bx_debugger.UCmem_journal.size == 0) {
      // start off point head & tail at same element
      bx_debugger.UCmem_journal.head = 0;
      tail = bx_debugger.UCmem_journal.tail = 0;
      }
    else {
      tail = bx_debugger.UCmem_journal.tail + 1;
      }
    if (tail >= BX_DBG_UCMEM_JOURNAL_SIZE) {
      fprintf(stderr, "dbg_ucmem_read: journal wrapped.\n");
      bx_dbg_exit(0);
      }

    value = bx_devices.vga->mem_read(addr);
    bx_dbg_ucmem_report(addr, 1, BX_READ, value);
    bx_debugger.UCmem_journal.element[tail].op    = BX_READ;
    bx_debugger.UCmem_journal.element[tail].len   = 1;
    bx_debugger.UCmem_journal.element[tail].addr  = addr;
    bx_debugger.UCmem_journal.element[tail].value = value;
    if (bx_debugger.UCmem_journal.size)
      bx_debugger.UCmem_journal.tail++;
    bx_debugger.UCmem_journal.size++;

if (doit)
fprintf(stderr, "MASTER UCR: head:%u tail%u size:%u\n",
  bx_debugger.UCmem_journal.head,
  bx_debugger.UCmem_journal.tail,
  bx_debugger.UCmem_journal.size);

    return(value);
    }
  else {
    if (bx_debugger.UCmem_journal.size == 0) {
      fprintf(stderr, "Error: ucmem_read: journal empty.\n");
      return(0xff);
      }
    head = bx_debugger.UCmem_journal.head;
    value = bx_debugger.UCmem_journal.element[head].value;

    if ((bx_debugger.UCmem_journal.element[head].op != BX_READ) ||
        (bx_debugger.UCmem_journal.element[head].len != 1) ||
        (bx_debugger.UCmem_journal.element[head].addr != addr)) {

      fprintf(stderr, "Error: ucmem_read: out of sync with journal.\n");
      fprintf(stderr, "Error: master: op=%1s len=%u addr=0x%x val=0x%x\n",
        (bx_debugger.UCmem_journal.element[head].op==BX_READ) ? "W" : "R",
        (unsigned) bx_debugger.UCmem_journal.element[head].len,
        (unsigned) bx_debugger.UCmem_journal.element[head].addr,
        (unsigned) bx_debugger.UCmem_journal.element[head].value);
      fprintf(stderr, "Error: slave:  op=W len=%u addr=0x%x val=0x%x\n",
        (unsigned) 1, (unsigned) addr, (unsigned) value);
      return(0xff);
      }
    // slave UCmem op in sync with journaled master op, delete this entry
    bx_debugger.UCmem_journal.head++;
    bx_debugger.UCmem_journal.size--;
    return(value);
    }
}

  void
bx_dbg_ucmem_write(Bit32u addr, Bit8u value)
{
  unsigned tail, head;

  if ( bx_debugger.master_slave_mode == BX_DBG_MASTER_MODE ) {
    if (bx_debugger.UCmem_journal.size >= BX_DBG_UCMEM_JOURNAL_SIZE) {
      fprintf(stderr, "dbg_ucmem_write: journal full.\n");
      bx_dbg_exit(0);
      }

    if (bx_debugger.UCmem_journal.size == 0) {
      // start off point head & tail at same element
      bx_debugger.UCmem_journal.head = 0;
      tail = bx_debugger.UCmem_journal.tail = 0;
      }
    else {
      tail = bx_debugger.UCmem_journal.tail + 1;
      }
    if (tail >= BX_DBG_UCMEM_JOURNAL_SIZE) {
      fprintf(stderr, "dbg_ucmem_write: journal wrapped.\n");
      bx_dbg_exit(0);
      }

    bx_debugger.UCmem_journal.element[tail].op    = BX_WRITE;
    bx_debugger.UCmem_journal.element[tail].len   = 1;
    bx_debugger.UCmem_journal.element[tail].addr  = addr;
    bx_debugger.UCmem_journal.element[tail].value = value;

    if (bx_debugger.UCmem_journal.size)
      bx_debugger.UCmem_journal.tail++;
    bx_debugger.UCmem_journal.size++;
    bx_devices.vga->mem_write(addr, value);
    bx_dbg_ucmem_report(addr, 1, BX_WRITE, value);
    }
  else {
    if (bx_debugger.UCmem_journal.size == 0) {
      fprintf(stderr, "Error: ucmem_write: journal empty.\n");
      return;
      }
    head = bx_debugger.UCmem_journal.head;

    if ((bx_debugger.UCmem_journal.element[head].op != BX_WRITE) ||
        (bx_debugger.UCmem_journal.element[head].len != 1) ||
        (bx_debugger.UCmem_journal.element[head].addr != addr) ||
        (bx_debugger.UCmem_journal.element[head].value != value) ) {
      fprintf(stderr, "Error: ucmem_write: out of sync with journal.\n");
      fprintf(stderr, "Error: master: op=%1s len=%u addr=0x%x val=0x%x\n",
        (bx_debugger.UCmem_journal.element[head].op==BX_WRITE) ? "W" : "R",
        (unsigned) bx_debugger.UCmem_journal.element[head].len,
        (unsigned) bx_debugger.UCmem_journal.element[head].addr,
        (unsigned) bx_debugger.UCmem_journal.element[head].value);
      fprintf(stderr, "Error: slave:  op=W len=%u addr=0x%x val=0x%x\n",
        (unsigned) 1, (unsigned) addr, (unsigned) value);
      return;
      }
    // slave UCmem op in sync with journaled master op, delete this entry
    bx_debugger.UCmem_journal.head++;
    bx_debugger.UCmem_journal.size--;
    }
}

  void
bx_dbg_async_pin_request(unsigned what, Boolean val)
{
  // Request from IO devices for change in pin external to CPU.
  // This is pended until CPU ack's with bx_dbg_async_pin_ack().

  if (bx_debugger.master_slave_mode != BX_DBG_MASTER_MODE) {
    fprintf(stderr, "Error: dbg_async_pin_request not in master mode.\n");
    bx_dbg_exit(1);
    }

  switch (what) {
    case BX_DBG_ASYNC_PENDING_A20:
      // Q pending status
      bx_guard.async_changes_pending.which |= BX_DBG_ASYNC_PENDING_A20;
      bx_guard.async_changes_pending.a20    = val;
      return;
      break;

    case BX_DBG_ASYNC_PENDING_RESET:
    case BX_DBG_ASYNC_PENDING_NMI:
    default:
      fprintf(stderr, "Error: set_async_pin: unhandled case.\n");
      bx_dbg_exit(1);
    }
}


  void
bx_dbg_async_pin_ack(unsigned what, Boolean val)
{
  // Acknowledgement from master simulator for pending change in pin
  // external to CPU.

  if (bx_debugger.master_slave_mode != BX_DBG_MASTER_MODE) {
    fprintf(stderr, "Error: dbg_async_pin_ack: not master mode.\n");
    bx_dbg_exit(1);
    }

  switch (what) {
    case BX_DBG_ASYNC_PENDING_A20:
      // get rid of pending status
      bx_guard.async_changes_pending.which &= ~BX_DBG_ASYNC_PENDING_A20;
      // notify pc_system of change
      bx_pc_system.set_enable_a20(val);
      if (bx_dbg_callback[bx_debugger.master].set_A20)
        bx_dbg_callback[bx_debugger.master].set_A20(val);
      bx_dbg_journal_a20_event(val);
      return;
      break;

    case BX_DBG_ASYNC_PENDING_RESET:
    case BX_DBG_ASYNC_PENDING_NMI:
    default:
      fprintf(stderr, "Error: set_async_pin: unhandled case.\n");
      bx_dbg_exit(1);
    }
}




  Bit32u
bx_dbg_inp(Bit16u addr, unsigned len)
{
  Bit32u value;
  unsigned tail, head;

  if ( bx_debugger.master_slave_mode == BX_DBG_MASTER_MODE ) {
    if (bx_debugger.IO_journal.size >= BX_DBG_IO_JOURNAL_SIZE) {
      fprintf(stderr, "dbg_inp: journal full.\n");
      bx_dbg_exit(0);
      }

    if (bx_debugger.IO_journal.size == 0) {
      // start off point head & tail at same element
      bx_debugger.IO_journal.head = 0;
      tail = bx_debugger.IO_journal.tail = 0;
      }
    else {
      tail = bx_debugger.IO_journal.tail + 1;
      }
    if (tail >= BX_DBG_IO_JOURNAL_SIZE) {
      fprintf(stderr, "dbg_inp: journal wrapped.\n");
      bx_dbg_exit(0);
      }

    value = bx_pc_system.inp(addr, len);
    bx_debugger.IO_journal.element[tail].op    = BX_READ;
    bx_debugger.IO_journal.element[tail].len   = (Bit8u) len;
    bx_debugger.IO_journal.element[tail].addr  = addr;
    bx_debugger.IO_journal.element[tail].value = value;
    if (bx_debugger.IO_journal.size)
      bx_debugger.IO_journal.tail++;
    bx_debugger.IO_journal.size++;
//fprintf(stderr, "MASTER IN: head:%u tail%u size:%u\n",
//  bx_debugger.IO_journal.head,
//  bx_debugger.IO_journal.tail,
//  bx_debugger.IO_journal.size);
    return(value);
    }
  else {
    if (bx_debugger.IO_journal.size == 0) {
      fprintf(stderr, "Error: dbg_inp: journal empty.\n");
      return(0xffffffff);
      }
    head = bx_debugger.IO_journal.head;
    value = bx_debugger.IO_journal.element[head].value;

    if ((bx_debugger.IO_journal.element[head].op != BX_READ) ||
        (bx_debugger.IO_journal.element[head].len != len) ||
        (bx_debugger.IO_journal.element[head].addr != addr) ) {
      fprintf(stderr, "Error: dbg_inp: out of sync with journal.\n");
      fprintf(stderr, "Error: master: op=%3s len=%u addr=0x%x\n",
        (bx_debugger.IO_journal.element[head].op==BX_WRITE) ? "OUT" : "IN",
        (unsigned) bx_debugger.IO_journal.element[head].len,
        (unsigned) bx_debugger.IO_journal.element[head].addr);
      fprintf(stderr, "Error: slave:  op=OUT len=%u addr=0x%x\n",
        (unsigned) len, (unsigned) addr);
      return(0xffffffff);
      }
    // slave IO op in sync with journaled master op, delete this entry
    bx_debugger.IO_journal.head++;
    bx_debugger.IO_journal.size--;
// fprintf(stderr, "SLAVE   IN: head:%u tail%u size:%u\n",
//   bx_debugger.IO_journal.head,
//   bx_debugger.IO_journal.tail,
//   bx_debugger.IO_journal.size);
    return(value);
    }
}

  void
bx_dbg_outp(Bit16u addr, Bit32u value, unsigned len)
{
  unsigned tail, head;

  if ( bx_debugger.master_slave_mode == BX_DBG_MASTER_MODE ) {
    if (bx_debugger.IO_journal.size >= BX_DBG_IO_JOURNAL_SIZE) {
      fprintf(stderr, "dbg_outp: IO journal full.\n");
      bx_dbg_exit(0);
      }

    if (bx_debugger.IO_journal.size == 0) {
      // start off point head & tail at same element
      bx_debugger.IO_journal.head = 0;
      tail = bx_debugger.IO_journal.tail = 0;
      }
    else {
      tail = bx_debugger.IO_journal.tail + 1;
      }
    if (tail >= BX_DBG_IO_JOURNAL_SIZE) {
      fprintf(stderr, "dbg_outp: IO journal wrapped.\n");
      bx_dbg_exit(0);
      }

    bx_debugger.IO_journal.element[tail].op    = BX_WRITE;
    bx_debugger.IO_journal.element[tail].len   = (Bit8u) len;
    bx_debugger.IO_journal.element[tail].addr  = addr;
    bx_debugger.IO_journal.element[tail].value = value;
    if (bx_debugger.IO_journal.size)
      bx_debugger.IO_journal.tail++;
    bx_debugger.IO_journal.size++;
    bx_pc_system.outp(addr, value, len);
if (doit)
fprintf(stderr, "master: IO journal size now %u\n", bx_debugger.IO_journal.size);
    }
  else {
    if (bx_debugger.IO_journal.size == 0) {
      fprintf(stderr, "Error: dbg_outp: journal empty.\n");
      return;
      }
    head = bx_debugger.IO_journal.head;

    if ((bx_debugger.IO_journal.element[head].op != BX_WRITE) ||
        (bx_debugger.IO_journal.element[head].len != len) ||
        (bx_debugger.IO_journal.element[head].addr != addr) ||
        (bx_debugger.IO_journal.element[head].value != value) ) {
      fprintf(stderr, "Error: dbg_outp: out of sync with journal.\n");
      fprintf(stderr, "Error: master: op=%3s len=%u addr=0x%x val=0x%x\n",
        (bx_debugger.IO_journal.element[head].op==BX_WRITE) ? "OUT" : "IN",
        (unsigned) bx_debugger.IO_journal.element[head].len,
        (unsigned) bx_debugger.IO_journal.element[head].addr,
        (unsigned) bx_debugger.IO_journal.element[head].value);
      fprintf(stderr, "Error: slave:  op=OUT len=%u addr=0x%x val=0x%x\n",
        (unsigned) len, (unsigned) addr, (unsigned) value);
      return;
      }
    // slave IO op in sync with journaled master op, delete this entry
    bx_debugger.IO_journal.head++;
    bx_debugger.IO_journal.size--;
if (doit)
fprintf(stderr, "slave: IO journal size now %u\n", bx_debugger.IO_journal.size);
    }
}

  void
bx_dbg_raise_HLDA(void)
{
  fprintf(stderr, "dbg_HLDA called\n");
  bx_dbg_exit(0);
}

  Bit8u
bx_dbg_IAC(void)
{
  // Convience routine.  bochs skips this, and calls the PIC code
  // directly.  This is for other simulators to interface to the
  // the PIC code.
  unsigned iac;

  iac = bx_devices.pic->IAC();
  return(iac);
}

  void
bx_dbg_set_INTR(Boolean b)
{
  if ( bx_debugger.master_slave_mode == BX_DBG_SLAVE_MODE ) {
    fprintf(stderr, "Error: set_INTR in slave mode.\n");
    bx_dbg_exit(1);
    }

  bx_pc_system.INTR = b;
  bx_dbg_callback[bx_debugger.master].set_INTR(b);
}

#endif  // #if (BX_NUM_SIMULATORS >= 2)
