// Copyright Notice
// ================
// BOCHS is Copyright 1994-1999 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
//
//     EMail:    bochs@world.std.com

// macintosh.cc -- bochs GUI file for the Macintosh
// written by David Batterham <drbatter@socs.uts.edu.au> or <drbatter@yahoo.com>

// --- BOCHS INCLUDES ---
#include "bochs.h"
#include "icon_bochs.h"
#include "font/vga.bitmap.h"

// --- MAC OS INCLUDES ---
#include <Quickdraw.h>
#include <QDOffscreen.h>
#include <Palettes.h>
#include <Windows.h>
#include <Memory.h>
#include <Events.h>
#include <TextUtils.h>
#include <Dialogs.h>
#include <LowMem.h>
#include <Disks.h>
#include <PP_KeyCodes.h>
#include <SIOUX.h>

// --- CONSTANTS ---
#define PIXEL_SIZE 1	// "Zoom" factor for the VGA display; increase this to make the
											// display look bigger (but chunkier)
											
#define SLEEP_TIME 15 // Number of ticks to surrender the processor during a WaitNextEvent()

#define FONT_WIDTH	8
#define FONT_HEIGHT	16

#define WINBITMAP(w)		(((GrafPtr)(w))->portBits)

#define ASCII_1_MASK 0x00FF0000
#define ASCII_2_MASK 0x000000FF

const RGBColor	black = {0, 0, 0};
const RGBColor	medGrey = {0xCCCC, 0xCCCC, 0xCCCC};
const RGBColor	ltGrey = {0xEEEE, 0xEEEE, 0xEEEE};

// --- GLOBALS ---
WindowPtr			win, toolwin;
bx_gui_c			*thisGUI;
SInt16				gOldMBarHeight;
RgnHandle			mBarRgn, cnrRgn;
unsigned			prev_block_cursor_x = 0, prev_block_cursor_y = 0, mouse_button_state = 0;
CTabHandle		gCTable;
PixMapHandle	gTile;
PixMapHandle	vgafont[256];
Rect					srcTextRect, srcTileRect;
Point					prevPt={0,0};
Boolean				respondToDisk = false;
UInt32				gTransState=0;
Ptr						KCHR;

// --- HEADERBAR STUFF ---
int 					numPixMaps = 0, toolPixMaps = 0;
unsigned			bx_bitmap_left_xorigin = 0; // pixels from left
unsigned			bx_bitmap_right_xorigin = 0; // pixels from right
PixMapHandle	bx_pixmap[BX_MAX_PIXMAPS];

struct {
  PixMapHandle pm;
  unsigned xdim;
  unsigned ydim;
  unsigned xorigin;
  unsigned yorigin;
  unsigned alignment;
  void (*f)(void);
} bx_tool_pixmap[BX_MAX_PIXMAPS];

// --- UTILITY FUNCTION PROTOTYPES ---
void HandleKey(Bit32u message, Bit32u keyState, EventModifiers mods);
void HandleToolClick(Point where);
void HandleClick(Point where);
void UpdateRgn(RgnHandle rgn);
void HideMenubar(void);
void ShowMenubar(void);
void CreateKeyMap(void);
void CreateVGAFont(void);
PixMapHandle CreatePixMap(unsigned left, unsigned top, unsigned width,
	unsigned height, unsigned depth, CTabHandle clut);
void SetCTabEntry(CTabHandle ctab, int index, unsigned r, unsigned g, unsigned b);
unsigned char reverse_bitorder(unsigned char);

// ::SPECIFIC_INIT()
//
// Called from gui.cc, once upon program startup, to allow for the
// specific GUI code (X11, BeOS, ...) to be initialized.
//
// th: a 'this' pointer to the gui class.  If a function external to the
//     class needs access, store this pointer and use later.
// argc, argv: not used right now, but the intention is to pass native GUI
//     specific options from the command line.  (X11 options, BeOS options,...)
//
// tilewidth, tileheight: for optimization, graphics_tile_update() passes
//     only updated regions of the screen to the gui code to be redrawn.
//     These define the dimensions of a region (tile).
// headerbar_y:  A headerbar (toolbar) is display on the top of the
//     VGA window, showing floppy status, and other information.  It
//     always assumes the width of the current VGA mode width, but
//     it's height is defined by this parameter.

void bx_gui_c::specific_init(bx_gui_c *th, int argc, char **argv, unsigned tilewidth, unsigned tileheight,
                     unsigned headerbar_y)
{
  Rect					winRect, toolRect;
  unsigned char	*wTitle = "\pMacBochs - use Cmd-B to show/hide this tool window";
	
  InitGraf(&qd.thePort);
  InitWindows();
  InitMenus();
  InitDialogs(nil);
  InitCursor();
  MaxApplZone();
  // Initialise the toolbox

  thisGUI = th;

  SetRect(&srcTextRect, 0, 0, FONT_WIDTH-1, FONT_HEIGHT-1);
  SetRect(&srcTileRect, 0, 0, tilewidth, tileheight);

  CreateKeyMap();
  gCTable = GetCTable(8);
  CreateVGAFont();
  gTile = CreatePixMap(0, 0, tilewidth, tileheight, 8, gCTable);

  SetRect(&winRect, 0, 0, qd.screenBits.bounds.right,
  	qd.screenBits.bounds.bottom);
  win = NewWindow(NULL, &winRect, (unsigned char *) wTitle, true, plainDBox,
  	(WindowPtr)-1, false, 0);
  // Create a full-screen window for the Bochs display

  SetRect(&toolRect, 2, qd.screenBits.bounds.bottom-headerbar_y-4, qd.screenBits.bounds.right-4,
  	qd.screenBits.bounds.bottom-2);
  toolwin = NewCWindow(NULL, &toolRect, (unsigned char *) wTitle, true, floatProc,
  	(WindowPtr)-1, false, 0);
  // Create a moveable tool window for the "headerbar"

  HideMenubar();
  // Hide the menu bar

  SetPort(win);

  SIOUXSettings.autocloseonquit = true;
  SIOUXSettings.asktosaveonclose = false;

  UNUSED(argc);
  UNUSED(argv);

  // UNUSED(bochs_icon_bits);
  // global variable

  //if (bx_options.private_colormap) {
  //  fprintf(stderr, "WARNING: NOGUI: private_colormap option ignored.\n");
  //  }

}

// HandleKey()
//
// Handles keyboard-related events.

void HandleKey(Bit32u message, Bit32u keyState, EventModifiers mods)
{
	int	key;
	UInt32 trans;
	
	key = message & charCodeMask;
	
	if (mods & cmdKey)
	{
		switch (key)
		{
			case 'B':
			case 'b':
				if (IsWindowVisible(toolwin))
					HideWindow(toolwin);
				else
				{
					ShowWindow(toolwin);
					BringToFront(toolwin);
					SelectWindow(toolwin);
					thisGUI->show_headerbar();
				}
				break;
				
			case 'E':
			case 'e':
				DiskEject(1);
				break;
				
			case 'Q':
			case 'q':
				bx_panic("User terminated");
				break;
		}
		return;
	}
	
	if (mods & shiftKey)
	{
		bx_devices.keyboard->gen_scancode(BX_KEY_SHIFT_L | BX_KEY_PRESSED);
		bx_printf("Shift-");
	}
	if (mods & controlKey)
	{
		bx_devices.keyboard->gen_scancode(BX_KEY_CTRL_L | BX_KEY_PRESSED);
		bx_printf("Ctrl-");
	}
	if (mods & optionKey)
	{
		bx_devices.keyboard->gen_scancode(BX_KEY_ALT_L | BX_KEY_PRESSED);
		bx_printf("Alt-");
	}
	
	key = (message & keyCodeMask) >> 8;
	
	trans = KeyTranslate(KCHR, key, &gTransState);
	
	// KeyTranslate maps Mac virtual key codes to any type of character code
	// you like (in this case, Bochs key codes). Much nicer than a huge switch
	// statement!
	
	//key = (trans & ASCII_1_MASK); // >> 16;
	
	if (trans > 0)
		bx_devices.keyboard->gen_scancode(trans | keyState);
		
	if (mods & shiftKey)
	{
		bx_devices.keyboard->gen_scancode(BX_KEY_SHIFT_L | BX_KEY_RELEASED);
		bx_printf("Shift-");
	}
	if (mods & controlKey)
	{
		bx_devices.keyboard->gen_scancode(BX_KEY_CTRL_L | BX_KEY_RELEASED);
		bx_printf("Ctrl-");
	}
	if (mods & optionKey)
	{
		bx_devices.keyboard->gen_scancode(BX_KEY_ALT_L | BX_KEY_RELEASED);
		bx_printf("Alt-");
	}
		
}

// HandleToolClick()
//
// Handles mouse clicks in the Bochs tool window

void HandleToolClick(Point where)
{
	unsigned i;
	int xorigin;
	Rect bounds;
	
	SetPort(toolwin);
	GlobalToLocal(&where);
	for (i=0; i<toolPixMaps; i++)
	{
		if (bx_tool_pixmap[i].alignment == BX_GRAVITY_LEFT)
			xorigin = bx_tool_pixmap[i].xorigin;
		else
			xorigin = toolwin->portRect.right - bx_tool_pixmap[i].xorigin;
		SetRect(&bounds, xorigin, 0, xorigin+(**bx_tool_pixmap[i].pm).bounds.right,
			toolwin->portRect.bottom);
		if (PtInRect(where, &bounds))
			bx_tool_pixmap[i].f();
	}
}

// HandleClick()
//
// Handles mouse click events.

void HandleClick(Point where)
{
	short part;
	WindowPtr whichWindow;
	
	part = FindWindow(where, &whichWindow);
	
	switch(part)
	{
		case inDrag:
			DragWindow(whichWindow, where, &qd.screenBits.bounds);
			break;
			
		case inContent:
			if (whichWindow == win)
				mouse_button_state |= 0x01;
			else if (whichWindow == toolwin)
				HandleToolClick(where);
			break;
				
		default:
			break;
	}

}

// ::HANDLE_EVENTS()
//
// Called periodically (vga_update_interval in .bochsrc) so the
// the gui code can poll for keyboard, mouse, and other
// relevant events.

void bx_gui_c::handle_events(void)
{
	EventRecord	event;
	Point	mousePt;
	int dx, dy;
		
	if (WaitNextEvent(everyEvent, &event, SLEEP_TIME, NULL))
	{
		switch(event.what)
		{
			case mouseDown:
				HandleClick(event.where);
				mouse_button_state |= 0x01;
				break;
				
			case mouseUp:
				mouse_button_state &= ~0x01;
				break;
				
			case keyDown:
			case autoKey:
				HandleKey(event.message, BX_KEY_PRESSED, event.modifiers);
				break;
				
			case keyUp:
				HandleKey(event.message, BX_KEY_RELEASED, event.modifiers);
				break;
				
			//case updateEvt:
				//if ((WindowPtr)event.message == win)
				//{
					//BeginUpdate(win);
					//FillRgn(win->visRgn, &qd.black);
					//EndUpdate(win);
				//}
				
			case diskEvt:
				if (respondToDisk)
					floppyA_handler();
				else
					respondToDisk = true;
				
			default:
				break;
		}
	}
	
	GetMouse(&mousePt);
	if (!EqualPt(mousePt, prevPt))
	{
		dx = mousePt.h - prevPt.h;
		dy = mousePt.v - prevPt.v;
		
		bx_devices.keyboard->mouse_motion(dx, dy, mouse_button_state);
	}
	prevPt = mousePt;

}


// ::FLUSH()
//
// Called periodically, requesting that the gui code flush all pending
// screen update requests.

void bx_gui_c::flush(void)
{
	BeginUpdate(win);
	EndUpdate(win);
	BeginUpdate(toolwin);
	EndUpdate(toolwin);
	// an opportunity to make the Window Manager happy.
}


// ::CLEAR_SCREEN()
//
// Called to request that the VGA region is cleared.  Don't
// clear the area that defines the headerbar.

void bx_gui_c::clear_screen(void)
{
  Rect	sRect;

  SetPort(win);
	SetRect(&sRect, 0, 0, win->portRect.right, win->portRect.bottom);
	FillRect(&sRect, &qd.black);
}



// ::TEXT_UPDATE()
//
// Called in a VGA text mode, to update the screen with
// new content.
//
// old_text: array of character/attributes making up the contents
//           of the screen from the last call.  See below
// new_text: array of character/attributes making up the current
//           contents, which should now be displayed.  See below
//
// format of old_text & new_text: each is 4000 bytes long.
//     This represents 80 characters wide by 25 high, with
//     each character being 2 bytes.  The first by is the
//     character value, the second is the attribute byte.
//     I currently don't handle the attribute byte.
//
// cursor_x: new x location of cursor
// cursor_y: new y location of cursor

void bx_gui_c::text_update(Bit8u *old_text, Bit8u *new_text,
                      unsigned long cursor_x, unsigned long cursor_y)
{
	int						i;
	unsigned char	achar;
	int						x, y;
	Rect					destRect;
	
	SetPort(win);	
	if ( (prev_block_cursor_y*80 + prev_block_cursor_x) < 2000)
	{
	  achar = new_text[(prev_block_cursor_y*80 + prev_block_cursor_x)*2];
	  SetRect(&destRect, prev_block_cursor_x*(FONT_WIDTH-1),
	  	prev_block_cursor_y*(FONT_HEIGHT-1),
	  	prev_block_cursor_x*(FONT_WIDTH-1)+(FONT_WIDTH-1),
	  	prev_block_cursor_y*(FONT_HEIGHT-1)+(FONT_HEIGHT-1));
	  CopyBits( & (** ((BitMapHandle)(vgafont[achar])) ), &WINBITMAP(win),
  			&srcTextRect, &destRect, srcCopy, NULL);
  }
	
  for (i=0; i<4000; i+=2)
  {
    if (new_text[i] != old_text[i])
    {

      achar = new_text[i];

      x = ((i/2) % 80);
      y = ((i/2) / 80);

      SetRect(&destRect, x*(FONT_WIDTH-1), y*(FONT_HEIGHT-1),
      	x*(FONT_WIDTH-1)+FONT_WIDTH-1, y*(FONT_HEIGHT-1)+FONT_HEIGHT-1);

      CopyBits( & (** ((BitMapHandle)(vgafont[achar])) ), &WINBITMAP(win),
  			&srcTextRect, &destRect, srcCopy, NULL);
    }
  }

  prev_block_cursor_x = cursor_x;
  prev_block_cursor_y = cursor_y;

  // now draw character at new block cursor location in reverse
  if ((cursor_y*80 + cursor_x) < 2000 )
  {
  	achar = new_text[(cursor_y*80 + cursor_x)*2];
  	SetRect(&destRect, cursor_x*(FONT_WIDTH-1),
	  	cursor_y*(FONT_HEIGHT-1),
	  	cursor_x*(FONT_WIDTH-1)+(FONT_WIDTH-1),
	  	cursor_y*(FONT_HEIGHT-1)+(FONT_HEIGHT-1));
	  			
	  CopyBits( & (** ((BitMapHandle)(vgafont[achar])) ), &WINBITMAP(win),
  		&srcTextRect, &destRect, notSrcCopy, NULL);
  }

}


// ::PALETTE_CHANGE()
//
// Allocate a color in the native GUI, for this color, and put
// it in the colormap location 'index'.
// returns: 0=no screen update needed (color map change has direct effect)
//          1=screen updated needed (redraw using current colormap)

Boolean bx_gui_c::palette_change(unsigned index, unsigned red, unsigned green, unsigned blue)
{

	red *= 0x100;
	green *= 0x100;
	blue *=0x100;
	
	// For some reason the RGB colors passed by Bochs are too dark for the Mac;
	// multiply them by 0x100 to get the correct colours.

  SetCTabEntry (gCTable, index, red, green, blue);
  SetEntries(index, index, (**gCTable).ctTable);

  return(0);
}


// ::GRAPHICS_TILE_UPDATE()
//
// Called to request that a tile of graphics be drawn to the
// screen, since info in this region has changed.
//
// tile: array of 8bit values representing a block of pixels with
//       dimension equal to the 'tilewidth' & 'tileheight' parameters to
//       ::specific_init().  Each value specifies an index into the
//       array of colors you allocated for ::palette_change()
// x0: x origin of tile
// y0: y origin of tile
//
// note: origin of tile and of window based on (0,0) being in the upper
//       left of the window.

void bx_gui_c::graphics_tile_update(Bit8u *tile, unsigned x0, unsigned y0)
{
  Rect					destRect;

  SetPort(win);
  SetRect(&destRect, x0*PIXEL_SIZE,
  	y0 * PIXEL_SIZE,
  	x0 * PIXEL_SIZE + (**gTile).bounds.right * PIXEL_SIZE,
  	y0 * PIXEL_SIZE + (**gTile).bounds.bottom * PIXEL_SIZE);
		
  (**gTile).baseAddr = (Ptr)tile;
	// Just set the baseAddr to point to the tile. We don't want to draw
	// each pixel to the screen manually.
			
	CopyBits( & (** ((BitMapHandle)gTile) ), &WINBITMAP(win),
		&srcTileRect, &destRect, srcCopy, NULL);
	// Copy the pixmap to the screen	

}



// ::DIMENSION_UPDATE()
//
// Called when the VGA mode changes it's X,Y dimensions.
// Resize the window to this size, but you need to add on
// the height of the headerbar to the Y value.
//
// x: new VGA x size
// y: new VGA y size (add headerbar_y parameter from ::specific_init().

void bx_gui_c::dimension_update(unsigned x, unsigned y)
{
//  SizeWindow(win, x, y + gToolHeight, FALSE);
//		gToolWidth = x;
	SizeWindow(toolwin, qd.screenBits.bounds.right-4, toolwin->portRect.bottom, FALSE);
	show_headerbar();

	UNUSED (x);
	UNUSED (y);
// I'm not currently changing the Bochs window size in response to
// dimension_update requests. Instead, I just keep the whole screen black.
// In future it would be nice to change the monitor resolution itself.
}


// ::CREATE_BITMAP()
//
// Create a monochrome bitmap of size 'xdim' by 'ydim', which will
// be drawn in the headerbar.  Return an integer ID to the bitmap,
// with which the bitmap can be referenced later.
//
// bmap: packed 8 pixels-per-byte bitmap.  The pixel order is:
//       bit0 is the left most pixel, bit7 is the right most pixel.
// xdim: x dimension of bitmap
// ydim: y dimension of bitmap

unsigned bx_gui_c::create_bitmap(const unsigned char *bmap, unsigned xdim, unsigned ydim)
{
	unsigned i;
	unsigned char *data;
	long row_bytes, bytecount;
	
  bx_pixmap[numPixMaps] = CreatePixMap(0, 0, xdim, ydim, 1, GetCTable(1));
  row_bytes = (**(bx_pixmap[numPixMaps])).rowBytes & 0x3FFF;
  bytecount = row_bytes * ((long) (ydim-1));
  data = (unsigned char *)NewPtrClear(bytecount);

  for (i=0; i<ydim * xdim/8; i++)
  	data[i] = reverse_bitorder(bmap[i]);

  (**(bx_pixmap[numPixMaps])).baseAddr = (Ptr)data;

  numPixMaps++;
  return(numPixMaps-1);
}


// ::HEADERBAR_BITMAP()
//
// Called to install a bitmap in the bochs headerbar (toolbar).
//
// bmap_id: will correspond to an ID returned from
//     ::create_bitmap().  'alignment' is either BX_GRAVITY_LEFT
//     or BX_GRAVITY_RIGHT, meaning install the bitmap in the next
//     available leftmost or rightmost space.
// alignment: is either BX_GRAVITY_LEFT or BX_GRAVITY_RIGHT,
//     meaning install the bitmap in the next
//     available leftmost or rightmost space.
// f: a 'C' function pointer to callback when the mouse is clicked in
//     the boundaries of this bitmap.

unsigned bx_gui_c::headerbar_bitmap(unsigned bmap_id, unsigned alignment, void (*f)(void))
{
	unsigned hb_index;
	
	toolPixMaps++;
	hb_index = toolPixMaps-1;
  bx_tool_pixmap[hb_index].pm = bx_pixmap[bmap_id];
  bx_tool_pixmap[hb_index].alignment = alignment;
  bx_tool_pixmap[hb_index].f = f;
  if (alignment == BX_GRAVITY_LEFT)
  {
  	bx_tool_pixmap[hb_index].xorigin = bx_bitmap_left_xorigin;
  	bx_tool_pixmap[hb_index].yorigin = 0;
  	bx_bitmap_left_xorigin += (**bx_pixmap[bmap_id]).bounds.right;
  }
  else
  {
    bx_bitmap_right_xorigin += (**bx_pixmap[bmap_id]).bounds.right;
    bx_tool_pixmap[hb_index].xorigin = bx_bitmap_right_xorigin;
    bx_tool_pixmap[hb_index].yorigin = 0;
  }
  return(hb_index);
}


// ::SHOW_HEADERBAR()
//
// Show (redraw) the current headerbar, which is composed of
// currently installed bitmaps.

void bx_gui_c::show_headerbar(void)
{
	Rect	destRect;
	int		i, xorigin;
	
	SetPort(toolwin);
	RGBForeColor(&medGrey);
	FillRect(&toolwin->portRect, &qd.black);
	for (i=0; i<toolPixMaps; i++)
	{
		if (bx_tool_pixmap[i].alignment == BX_GRAVITY_LEFT)
			xorigin = bx_tool_pixmap[i].xorigin;
		else
			xorigin = toolwin->portRect.right - bx_tool_pixmap[i].xorigin;
			
		SetRect(&destRect, xorigin, 0, xorigin+(**bx_tool_pixmap[i].pm).bounds.right,
			(**bx_tool_pixmap[i].pm).bounds.bottom);
		RGBForeColor(&ltGrey);
		CopyBits( & (** ((BitMapHandle)bx_tool_pixmap[i].pm) ), &WINBITMAP(toolwin),
			&(**bx_tool_pixmap[i].pm).bounds, &destRect, srcOr, NULL);
		
		if (true)
		{
			OffsetRect(&destRect, 2, 2);
			RGBForeColor(&black);
			CopyBits( & (** ((BitMapHandle)bx_tool_pixmap[i].pm) ), &WINBITMAP(toolwin),
				&(**bx_tool_pixmap[i].pm).bounds, &destRect, srcOr, NULL);
		}
			
		// Draw each toolbar icon with a subtle 3D look
		
	}

}


// ::REPLACE_BITMAP()
//
// Replace the bitmap installed in the headerbar ID slot 'hbar_id',
// with the one specified by 'bmap_id'.  'bmap_id' will have
// been generated by ::create_bitmap().  The old and new bitmap
// must be of the same size.  This allows the bitmap the user
// sees to change, when some action occurs.  For example when
// the user presses on the floppy icon, it then displays
// the ejected status.
//
// hbar_id: headerbar slot ID
// bmap_id: bitmap ID

void bx_gui_c::replace_bitmap(unsigned hbar_id, unsigned bmap_id)
{
	bx_tool_pixmap[hbar_id].pm = bx_pixmap[bmap_id];
	show_headerbar;
}


// ::EXIT()
//
// Called before bochs terminates, to allow for a graceful
// exit from the native GUI mechanism.

void bx_gui_c::exit(void)
{
	ShowMenubar(); // Make the menubar visible again
}

// UpdateRgn()
//
// Updates the screen after the menubar and round corners have been hidden

void UpdateRgn(RgnHandle rgn)
{
	WindowRef window;
	
	window = LMGetWindowList();
	PaintBehind(window, rgn);
	CalcVisBehind(window, rgn);
}

// HideMenubar()
//
// Hides the menubar (obviously)

void HideMenubar()
{
	Rect mBarRect;
	RgnHandle grayRgn;

	grayRgn = LMGetGrayRgn();
	gOldMBarHeight = GetMBarHeight();
	LMSetMBarHeight(0);
	mBarRgn = NewRgn();
	mBarRect = qd.screenBits.bounds;
	mBarRect.bottom = mBarRect.top + gOldMBarHeight;
	RectRgn(mBarRgn, &mBarRect);
	UnionRgn(grayRgn, mBarRgn, grayRgn);
	UpdateRgn(mBarRgn);
	
	cnrRgn = NewRgn();
	RectRgn(cnrRgn, &qd.screenBits.bounds);
	DiffRgn(cnrRgn, grayRgn, cnrRgn);
	UnionRgn(grayRgn, cnrRgn, grayRgn);
	UpdateRgn(mBarRgn);
}

// ShowMenubar()
//
// Makes the menubar visible again so other programs will display correctly.

void ShowMenubar()
{
	RgnHandle	grayRgn;
	GrafPtr		savePort;
	
	grayRgn = LMGetGrayRgn();
	LMSetMBarHeight(gOldMBarHeight);
	DiffRgn(grayRgn, mBarRgn, grayRgn);
	DisposeRgn(mBarRgn);
	DrawMenuBar();
	
	GetPort(&savePort);
	SetPort(LMGetWMgrPort());
	SetClip(cnrRgn);
	FillRgn(cnrRgn, &qd.black);
	SetPort(savePort);
	DiffRgn(grayRgn, cnrRgn, grayRgn);
	DisposeRgn(cnrRgn);
}

// CreateKeyMap()
//
// Create a KCHR data structure to map Mac virtual key codes to Bochs key codes

void CreateKeyMap(void)
{
	const unsigned char KCHRHeader [258] = {
		0,
		1
	};

	const unsigned char KCHRTable [130] = {
		0,
		1,
		BX_KEY_A,
		BX_KEY_S,
		BX_KEY_D,
		BX_KEY_F,
		BX_KEY_H,
		BX_KEY_G,
		BX_KEY_Z,
		BX_KEY_X,
		BX_KEY_C,
		BX_KEY_V,
		0,
		BX_KEY_B,
		BX_KEY_Q,
		BX_KEY_W,
		BX_KEY_E,
		BX_KEY_R,
		BX_KEY_Y,
		BX_KEY_T,
		BX_KEY_1,
		BX_KEY_2,
		BX_KEY_3,
		BX_KEY_4,
		BX_KEY_6,
		BX_KEY_5,
		BX_KEY_EQUALS,
		BX_KEY_9,
		BX_KEY_7,
		BX_KEY_MINUS,
		BX_KEY_8,
		BX_KEY_0,
		BX_KEY_RIGHT_BRACKET,
		BX_KEY_O,
		BX_KEY_U,
		BX_KEY_LEFT_BRACKET,
		BX_KEY_I,
		BX_KEY_P,
		BX_KEY_ENTER,
		BX_KEY_L,
		BX_KEY_J,
		BX_KEY_SINGLE_QUOTE,
		BX_KEY_K,
		BX_KEY_SEMICOLON,
		BX_KEY_BACKSLASH,
		BX_KEY_COMMA,
		BX_KEY_SLASH,
		BX_KEY_N,
		BX_KEY_M,
		BX_KEY_PERIOD,
		BX_KEY_TAB,
		BX_KEY_SPACE,
		BX_KEY_GRAVE,
		BX_KEY_BACKSPACE,
		0,
		BX_KEY_ESC,
		0, // 0x36
		0, // 0x37
		0, // 0x38
		0, // 0x39
		0, // 0x3A
		0, // 0x3B
		0, // 0x3C
		0, // 0x3D
		0, // 0x3E
		0, // 0x3F
		0, // 0x40
		BX_KEY_PERIOD, // KP_PERIOD
		0, // 0x42
		BX_KEY_KP_MULTIPLY,
		0, // 0x44
		BX_KEY_KP_ADD,
		0, // 0x46
		BX_KEY_KP_DELETE,
		0, // 0x48
		0, // 0x49
		0, // 0x4A
		BX_KEY_KP_DIVIDE,
		BX_KEY_KP_ENTER,
		0, // 0x4D
		BX_KEY_KP_SUBTRACT,
		0, // 0x4F
		0, // 0x50
		0, // 0x51 (kp equals)
		0, // 0x52 (kp 0)
		0, // 0x53 (kp 1)
		BX_KEY_KP_DOWN, // 0x54 (kp 2)
		0, // 0x55 (kp 3)
		BX_KEY_KP_LEFT, // 0x56 (kp 4)
		BX_KEY_KP_5,
		BX_KEY_KP_RIGHT, // 0x58 (kp 6)
		0, // 0x59 (kp 7)
		0, // 0x5A
		BX_KEY_KP_UP, // 0x5B (kp 8)
		0, // 0x5C (kp 9)
		0, // 0x5D
		0, // 0x5E
		0, // 0x5F
		BX_KEY_F5,
		BX_KEY_F6,
		BX_KEY_F7,
		BX_KEY_F3,
		BX_KEY_F8,
		BX_KEY_F9,
		0, // 0x66
		BX_KEY_F11,
		0, // 0x68
		0, // 0x69 (print screen)
		0, // 0x6A
		0, // 0x6B (scroll lock)
		0, // 0x6C
		BX_KEY_F10,
		0, // 0x6E
		BX_KEY_F12,
		0, // 0x70
		0, // 0x71 (pause)
		BX_KEY_INSERT,
		BX_KEY_HOME,
		BX_KEY_PAGE_UP,
		BX_KEY_DELETE,
		BX_KEY_F4,
		BX_KEY_END,
		BX_KEY_F2,
		BX_KEY_PAGE_DOWN,
		BX_KEY_F1,
		BX_KEY_LEFT,
		BX_KEY_RIGHT,
		BX_KEY_DOWN,
		BX_KEY_UP
	};

  KCHR = NewPtrClear(390);
  BlockMove(KCHRHeader, KCHR, sizeof(KCHRHeader));
  BlockMove(KCHRTable, Ptr(KCHR + sizeof(KCHRHeader)), sizeof(KCHRTable));
}

// CreateVGAFont()
//
// Create an array of PixMaps for the PC screen font

void CreateVGAFont(void)
{
	int						i, x;
	unsigned char *fontData, curPixel;
	long					row_bytes, bytecount;
	
  for (i=0; i<256; i++)
  {
    vgafont[i] = CreatePixMap(0, 0, FONT_WIDTH-1, FONT_HEIGHT-1, 1, GetCTable(1));
    row_bytes = (**(vgafont[i])).rowBytes & 0x3FFF;
	  bytecount = row_bytes * ((long) (FONT_HEIGHT-1));
    fontData = (unsigned char *)NewPtrClear(bytecount);

    for (x=0; x<16; x++)
    {
    	curPixel = ~(bx_vgafont[i].data[x]);
    	fontData[x*row_bytes] = reverse_bitorder(curPixel);
		}
    (**(vgafont[i])).baseAddr = Ptr(fontData);

  }
}

// CreatePixMap()
// Allocate a new pixmap handle and fill in the fields with appropriate
// values.

PixMapHandle CreatePixMap(unsigned left, unsigned top, unsigned width,
	unsigned height, unsigned depth, CTabHandle clut)
{
	PixMapHandle	pm;
	long					row_bytes;
	
	row_bytes = (((long) depth * ((long) width) + 31) >> 5) << 2;
	pm = NewPixMap();
	(**pm).bounds.left = left;
	(**pm).bounds.top = top;
	(**pm).bounds.right = left+width;
	(**pm).bounds.bottom = top+height;
	//(**pm).packType = 0;
	//(**pm).packSize = 0;
	//(**pm).pixelType = 0;
	(**pm).pixelSize = depth;
	//(**pm).cmpCount = 1;
	//(**pm).cmpSize = depth;
	(**pm).rowBytes = row_bytes | 0x8000;
	
  DisposeCTable((**pm).pmTable);
  (**pm).pmTable = clut;
  // Quickdraw allocates a new color table by default, but we want to
  // use one we created earlier.
  	
	return pm;
}

// SetCTabEntry()
// Changes the RGB color at the specified index in the color table

void SetCTabEntry(CTabHandle ctab, int index, unsigned r, unsigned g, unsigned b)
{

	(**ctab).ctTable[index].rgb.red = r;
	(**ctab).ctTable[index].rgb.green = g;
	(**ctab).ctTable[index].rgb.blue = b;
}

unsigned char reverse_bitorder(unsigned char b)
{
  unsigned char ret=0;
	
  for (unsigned i=0; i<8; i++)
  {
    ret |= (b & 0x01) << (7-i);
    b >>= 1;
  }
	
  return(ret);
}
